import math
from typing import TYPE_CHECKING, Self

from pydantic import BaseModel, PrivateAttr, model_validator

from hllrcon.data.maps import Orientation

if TYPE_CHECKING:
    from hllrcon.data.layers import Layer

GridPos = tuple[int, int]
GridArea = tuple[GridPos, GridPos]
WorldPos2D = tuple[float, float]
WorldPos3D = tuple[float, float, float]
WorldArea2D = tuple[WorldPos2D, WorldPos2D]


class Grid(BaseModel, frozen=True):
    scale: float
    offset: WorldPos2D
    size: GridArea

    @classmethod
    def large(
        cls,
        *,
        scale: float = 20000.0,
        offset: WorldPos2D = (0.0, 0.0),
        size: GridArea = ((-5, -5), (4, 4)),
    ) -> Self:
        return cls(
            scale=scale,
            offset=offset,
            size=size,
        )

    @classmethod
    def small(
        cls,
        *,
        scale: float = 13926.4,
        offset: WorldPos2D = (0.0, 0.0),
        size: GridArea = ((-5, -5), (4, 4)),
    ) -> Self:
        return cls(
            scale=scale,
            offset=offset,
            size=size,
        )

    def grid_to_world_from(self, grid_pos: GridPos) -> WorldPos2D:
        x = grid_pos[0] * self.scale + self.offset[0]
        y = grid_pos[1] * self.scale + self.offset[1]
        return (x, y)

    def grid_to_world_to(self, grid_pos: GridPos) -> WorldPos2D:
        x = (grid_pos[0] + 1) * self.scale + self.offset[0]
        y = (grid_pos[1] + 1) * self.scale + self.offset[1]
        return (x, y)

    def world_to_grid(self, world_pos: WorldPos2D) -> GridPos:
        x = (world_pos[0] - self.offset[0]) / self.scale
        y = (world_pos[1] - self.offset[1]) / self.scale
        return (math.floor(x), math.floor(y))


class GridPositionalModel(BaseModel, frozen=True):
    _layer: "Layer" = PrivateAttr()
    grid_from: GridPos
    grid_to: GridPos

    @model_validator(mode="after")
    def _validate_grid_from_to_order(self) -> Self:
        if self.grid_from[0] > self.grid_to[0] or self.grid_from[1] > self.grid_to[1]:
            msg = "grid_from must be smaller than grid_to"
            raise ValueError(msg)
        return self

    @property
    def from_(self) -> WorldPos2D:
        return self._layer.grid.grid_to_world_from(self.grid_from)

    @property
    def to(self) -> WorldPos2D:
        return self._layer.grid.grid_to_world_to(self.grid_to)

    @property
    def area(self) -> WorldArea2D:
        return (self.from_, self.to)

    def is_inside(self, world_pos: WorldPos2D) -> bool:
        x1, y1 = self.from_
        x2, y2 = self.to
        px, py = world_pos
        return x1 <= px <= x2 and y1 <= py <= y2


class Strongpoint(BaseModel, frozen=True):
    name: str
    center: WorldPos3D
    radius: float

    def is_inside(self, pos: tuple[float, float, float]) -> bool:
        """Check whether a given position is inside the strongpoint.

        Parameters
        ----------
        pos : tuple[float, float, float]
            The position to check.

        Returns
        -------
        bool
            True if the position is inside the strongpoint, False otherwise.

        """
        dx = pos[0] - self.center[0]
        dy = pos[1] - self.center[1]
        dz = pos[2] - self.center[2]
        return dx * dx + dy * dy + dz * dz <= self.radius * self.radius


class CaptureZone(GridPositionalModel, frozen=True):
    strongpoint: Strongpoint


class Sector(GridPositionalModel, frozen=True):
    capture_zones: list[CaptureZone]

    @classmethod
    def large_layout(
        cls,
        orientation: Orientation,
        strongpoints: tuple[
            tuple[Strongpoint, Strongpoint, Strongpoint],
            tuple[Strongpoint, Strongpoint, Strongpoint],
            tuple[Strongpoint, Strongpoint, Strongpoint],
            tuple[Strongpoint, Strongpoint, Strongpoint],
            tuple[Strongpoint, Strongpoint, Strongpoint],
        ],
    ) -> list[Self]:
        if orientation == Orientation.HORIZONTAL:

            def orient(x: int, y: int) -> tuple[int, int]:
                return x, y
        else:

            def orient(x: int, y: int) -> tuple[int, int]:
                return y, x

        return [
            cls(
                grid_from=orient(2 * i - 5, -3),
                grid_to=orient(2 * i - 4, 2),
                capture_zones=[
                    CaptureZone(
                        grid_from=orient(2 * i - 5, 2 * j - 3),
                        grid_to=orient(2 * i - 4, 2 * j - 2),
                        strongpoint=strongpoint,
                    )
                    for j, strongpoint in enumerate(sector)
                ],
            )
            for i, sector in enumerate(strongpoints)
        ]

    @classmethod
    def skirmish_layout(
        cls,
        orientation: Orientation,
        strongpoint: Strongpoint,
    ) -> list[Self]:
        if orientation == Orientation.HORIZONTAL:

            def orient(x: int, y: int) -> tuple[int, int]:
                return x, y
        else:

            def orient(x: int, y: int) -> tuple[int, int]:
                return y, x

        return [
            cls(
                grid_from=orient(-5, -4),
                grid_to=orient(-2, 3),
                capture_zones=[],
            ),
            cls(
                grid_from=orient(-1, -4),
                grid_to=orient(0, 3),
                capture_zones=[
                    CaptureZone(
                        grid_from=orient(-1, -1),
                        grid_to=orient(0, 0),
                        strongpoint=strongpoint,
                    ),
                ],
            ),
            cls(
                grid_from=orient(1, -4),
                grid_to=orient(4, 3),
                capture_zones=[],
            ),
        ]


SECTORS_CARENTAN_LARGE = Sector.large_layout(
    orientation=Orientation.HORIZONTAL,
    strongpoints=(
        (
            Strongpoint(
                name="Blactot",
                center=(-65543.41, -39731.965, 1359.9531),
                radius=5000.0,
            ),
            Strongpoint(
                name="502nd Start",
                center=(-67076.41, 4670.035, 123.953125),
                radius=5000.0,
            ),
            Strongpoint(
                name="Farm Ruins",
                center=(-68814.41, 37720.035, 365.95312),
                radius=5000.0,
            ),
        ),
        (
            Strongpoint(
                name="Pumping Station",
                center=(-36748.406, -29821.965, 146.95312),
                radius=5000.0,
            ),
            Strongpoint(
                name="Ruins",
                center=(-26183.406, 2343.0352, 101.953125),
                radius=3000.0,
            ),
            Strongpoint(
                name="Derailed Train",
                center=(-39381.406, 28975.035, 279.95312),
                radius=5000.0,
            ),
        ),
        (
            Strongpoint(
                name="Canal Crossing",
                center=(5892.5938, -39387.965, 279.95312),
                radius=5000.0,
            ),
            Strongpoint(
                name="Town Center",
                center=(1021.59375, -1021.96484, 104.953125),
                radius=5000.0,
            ),
            Strongpoint(
                name="Train Station",
                center=(246.59375, 27698.035, 176.95312),
                radius=5000.0,
            ),
        ),
        (
            Strongpoint(
                name="Customs",
                center=(40816.594, -34224.965, 279.95312),
                radius=5000.0,
            ),
            Strongpoint(
                name="Rail Crossing",
                center=(44171.594, -6296.965, 279.95312),
                radius=5000.0,
            ),
            Strongpoint(
                name="Mount Halais",
                center=(33828.594, 51343.035, 2518.9531),
                radius=3973.6313,
            ),
        ),
        (
            Strongpoint(
                name="Canal Locks",
                center=(66826.59, -26456.965, 279.95312),
                radius=5000.0,
            ),
            Strongpoint(
                name="Rail Causeway",
                center=(75611.59, 5968.035, 279.95312),
                radius=5495.63,
            ),
            Strongpoint(
                name="La Maison Des Ormes",
                center=(72222.59, 38476.035, 103.53516),
                radius=5000.0,
            ),
        ),
    ),
)

SECTORS_CARENTAN_SMALL = Sector.skirmish_layout(
    orientation=Orientation.HORIZONTAL,
    strongpoint=Strongpoint(
        name="Town Centre",
        center=(60.0, 160.0, -850.0),
        radius=8000.0,
    ),
)

SECTORS_DRIEL_LARGE = Sector.large_layout(
    orientation=Orientation.VERTICAL,
    strongpoints=(
        (
            Strongpoint(
                name="Oosterbeek Approach",
                center=(-36028.543, -79955.25, -412.13928),
                radius=6000.0,
            ),
            Strongpoint(
                name="Roseander Polder",
                center=(2809.0745, -78795.875, -159.22235),
                radius=6000.0,
            ),
            Strongpoint(
                name="Kasteel Rosande",
                center=(38371.14, -76418.7, 86.83746),
                radius=7000.0,
            ),
        ),
        (
            Strongpoint(
                name="Boatyard",
                center=(-38518.715, -33980.625, -205.24878),
                radius=7000.0,
            ),
            Strongpoint(
                name="Bridgeway",
                center=(3880.9673, -39449.43, -249.79703),
                radius=6000.0,
            ),
            Strongpoint(
                name="Rijn Banks",
                center=(39177.535, -42960.49, -309.9341),
                radius=7000.0,
            ),
        ),
        (
            Strongpoint(
                name="Brick Factory",
                center=(-39703.027, 6122.7656, -205.24878),
                radius=6000.0,
            ),
            Strongpoint(
                name="Railway Bridge",
                center=(2882.3755, -3877.2988, -205.24878),
                radius=9000.0,
            ),
            Strongpoint(
                name="Gun Emplacements",
                center=(43301.99, -2530.0012, -205.24878),
                radius=5500.0,
            ),
        ),
        (
            Strongpoint(
                name="Rietveld",
                center=(-40615.844, 40909.707, -375.4043),
                radius=6000.0,
            ),
            Strongpoint(
                name="South Railway",
                center=(3826.8418, 42206.754, -429.63922),
                radius=8000.0,
            ),
            Strongpoint(
                name="Middel Road",
                center=(41461.46, 38457.824, -205.24878),
                radius=6000.0,
            ),
        ),
        (
            Strongpoint(
                name="Orchards",
                center=(-39533.195, 77266.98, -329.62988),
                radius=8000.0,
            ),
            Strongpoint(
                name="Schaduwwolken Farm",
                center=(-2113.1738, 75816.06, -357.01),
                radius=6500.0,
            ),
            Strongpoint(
                name="Fields",
                center=(41461.46, 75453.516, -205.25464),
                radius=6000.0,
            ),
        ),
    ),
)

SECTORS_DRIEL_SMALL = Sector.skirmish_layout(
    orientation=Orientation.VERTICAL,
    strongpoint=Strongpoint(
        name="Underpass",
        center=(2600.0, -750.0, 450.0),
        radius=8000.0,
    ),
)

SECTORS_ELALAMEIN_LARGE = Sector.large_layout(
    orientation=Orientation.HORIZONTAL,
    strongpoints=(
        (
            Strongpoint(
                name="Vehicle Depot",
                center=(-68233.38, -37264.52, 968.23914),
                radius=6000.0,
            ),
            Strongpoint(
                name="Artillery Guns",
                center=(-71609.83, -8175.4565, -228.91907),
                radius=6000.0,
            ),
            Strongpoint(
                name="Miteiriya Ridge",
                center=(-79261.695, 36680.625, 1629.1664),
                radius=6000.0,
            ),
        ),
        (
            Strongpoint(
                name="Hamlet Ruins",
                center=(-37466.633, -37732.38, -1402.2225),
                radius=6000.0,
            ),
            Strongpoint(
                name="El Mreir",
                center=(-37776.816, -2887.5278, -1248.4254),
                radius=6000.0,
            ),
            Strongpoint(
                name="Watchtower",
                center=(-40818.59, 37838.586, 648.23755),
                radius=6000.0,
            ),
        ),
        (
            Strongpoint(
                name="Desert Rat Trenches",
                center=(4880.006, -40988.05, 831.8468),
                radius=6000.0,
            ),
            Strongpoint(
                name="Oasis",
                center=(-2900.921, -851.27783, -1248.4254),
                radius=6000.0,
            ),
            Strongpoint(
                name="Valley",
                center=(1970.4421, 35186.074, -787.2982),
                radius=8190.72,
            ),
        ),
        (
            Strongpoint(
                name="Fuel Depot",
                center=(43333.848, -35426.484, -1862.4927),
                radius=7000.0,
            ),
            Strongpoint(
                name="Airfield Command",
                center=(38495.92, -4155.8906, -1057.3627),
                radius=6000.0,
            ),
            Strongpoint(
                name="Airfield Hangars",
                center=(41085.367, 32927.33, -1248.4176),
                radius=8000.0,
            ),
        ),
        (
            Strongpoint(
                name="Cliffside Village",
                center=(68942.24, -39028.402, 550.42114),
                radius=6000.0,
            ),
            Strongpoint(
                name="Ambushed Convoy",
                center=(72480.45, -2526.4434, -1248.4176),
                radius=6000.0,
            ),
            Strongpoint(
                name="Quarry",
                center=(78760.73, 41540.4, -69.86389),
                radius=6000.0,
            ),
        ),
    ),
)

SECTORS_ELALAMEIN_SMALL = Sector.skirmish_layout(
    orientation=Orientation.HORIZONTAL,
    strongpoint=Strongpoint(
        name="Oasis",
        center=(-171.79688, -4065.439, -850.0),
        radius=6000.0,
    ),
)

SECTORS_ELSENBORNRIDGE_LARGE = Sector.large_layout(
    orientation=Orientation.VERTICAL,
    strongpoints=(
        (
            Strongpoint(
                name="99th Command Centre",
                center=(-39637.496, -67610.96, 5383.203),
                radius=7500.0,
            ),
            Strongpoint(
                name="Gun Battery",
                center=(420.0, -69376.0, 6308.203),
                radius=8000.0,
            ),
            Strongpoint(
                name="U.S. Camp",
                center=(50979.016, -67675.0, 6308.203),
                radius=7000.0,
            ),
        ),
        (
            Strongpoint(
                name="Elsenborn Ridge",
                center=(-30950.0, -41967.96, 6858.203),
                radius=7000.0,
            ),
            Strongpoint(
                name="Farahilde Farm",
                center=(10158.0, -30210.0, 5808.203),
                radius=8000.0,
            ),
            Strongpoint(
                name="Jensit Pillboxes",
                center=(49674.99, -28085.947, 5108.2026),
                radius=7000.0,
            ),
        ),
        (
            Strongpoint(
                name="Road To Elsenborn Ridge",
                center=(-40964.0, 3317.0, 6608.203),
                radius=8000.0,
            ),
            Strongpoint(
                name="Dug Out Tank",
                center=(-9124.0, 2404.0, 5608.203),
                radius=6000.0,
            ),
            Strongpoint(
                name="Checkpoint",
                center=(40444.914, 6529.1445, 1716.1997),
                radius=5000.0,
            ),
        ),
        (
            Strongpoint(
                name="Erelsdell Farmhouse",
                center=(-41672.0, 38246.0, 3633.2031),
                radius=8000.0,
            ),
            Strongpoint(
                name="AA Battery",
                center=(8607.678, 33127.07, 2796.5845),
                radius=7000.0,
            ),
            Strongpoint(
                name="Hinterburg",
                center=(39637.227, 39888.688, 3225.2002),
                radius=8000.0,
            ),
        ),
        (
            Strongpoint(
                name="Supply Cache",
                center=(-25666.0, 66300.0, 2862.2031),
                radius=5000.0,
            ),
            Strongpoint(
                name="Foxholes",
                center=(12223.855, 67172.27, -364.80078),
                radius=7000.0,
            ),
            Strongpoint(
                name="Fuel Depot",
                center=(38049.76, 70408.04, 2951.984),
                radius=7000.0,
            ),
        ),
    ),
)

SECTORS_ELSENBORNRIDGE_SMALL = Sector.skirmish_layout(
    orientation=Orientation.VERTICAL,
    strongpoint=Strongpoint(
        name="Dug Out Tank",
        center=(-8510.0, 1410.0, -279.18115),
        radius=5000.0,
    ),
)

SECTORS_FOY_LARGE = Sector.large_layout(
    orientation=Orientation.VERTICAL,
    strongpoints=(
        (
            Strongpoint(
                name="Road To Recogne",
                center=(-49755.0, -74340.0, -211.0),
                radius=2750.0,
            ),
            Strongpoint(
                name="Cobru Approach",
                center=(9952.0, -74787.0, -243.0),
                radius=3500.0,
            ),
            Strongpoint(
                name="Road To Noville",
                center=(38286.176, -76947.95, -243.0),
                radius=5343.75,
            ),
        ),
        (
            Strongpoint(
                name="Cobru Factory",
                center=(-29988.0, -44676.0, -890.0),
                radius=5500.0,
            ),
            Strongpoint(
                name="Foy",
                center=(-9586.0, -34052.0, -551.0),
                radius=3250.0,
            ),
            Strongpoint(
                name="Flak Battery",
                center=(45241.0, -39594.0, -964.0),
                radius=4000.0,
            ),
        ),
        (
            Strongpoint(
                name="West Bend",
                center=(-53153.0, -12966.0, -634.0),
                radius=5500.0,
            ),
            Strongpoint(
                name="Southern Edge",
                center=(-1114.0, 589.0, -102.0),
                radius=4738.37,
            ),
            Strongpoint(
                name="Dugout Barn",
                center=(46085.04, -4721.094, -1008.08936),
                radius=4139.884,
            ),
        ),
        (
            Strongpoint(
                name="N30 Highway",
                center=(-38407.0, 31775.0, -142.0),
                radius=6250.0,
            ),
            Strongpoint(
                name="Bizory Foy Road",
                center=(10035.0, 39390.0, -545.0),
                radius=3500.0,
            ),
            Strongpoint(
                name="Eastern Ourthe",
                center=(45845.0, 27822.0, -771.0),
                radius=4531.25,
            ),
        ),
        (
            Strongpoint(
                name="Road To Bastogne",
                center=(-52862.0, 63773.0, 112.0),
                radius=4000.0,
            ),
            Strongpoint(
                name="Bois Jacques",
                center=(-5582.0, 68237.0, 1106.0),
                radius=5000.0,
            ),
            Strongpoint(
                name="Forest Outskirts",
                center=(46279.0, 67141.0, 512.0),
                radius=5000.0,
            ),
        ),
    ),
)

SECTORS_HILL400_LARGE = Sector.large_layout(
    orientation=Orientation.HORIZONTAL,
    strongpoints=(
        (
            Strongpoint(
                name="Convoy Ambush",
                center=(-65875.18, -36966.816, 6207.824),
                radius=3000.0,
            ),
            Strongpoint(
                name="Federchecke Junction",
                center=(-65367.926, 2874.167, 10826.176),
                radius=4250.0,
            ),
            Strongpoint(
                name="Stuckchen Farm",
                center=(-63938.484, 42413.004, 8142.296),
                radius=3000.0,
            ),
        ),
        (
            Strongpoint(
                name="Roer River House",
                center=(-38405.066, -43380.766, -342.3706),
                radius=3000.0,
            ),
            Strongpoint(
                name="Bergstein Church",
                center=(-30580.357, 8420.501, 11575.116),
                radius=3000.0,
            ),
            Strongpoint(
                name="Kirchweg",
                center=(-41257.29, 31282.14, 8949.19),
                radius=3000.0,
            ),
        ),
        (
            Strongpoint(
                name="Flak Pits",
                center=(1384.4886, -33584.805, 8937.715),
                radius=3000.0,
            ),
            Strongpoint(
                name="Hill 400",
                center=(-1408.8995, 4698.0444, 17213.738),
                radius=5000.0,
            ),
            Strongpoint(
                name="Southern Approach",
                center=(948.21277, 25170.994, 12086.199),
                radius=3000.0,
            ),
        ),
        (
            Strongpoint(
                name="Eselsweg Junction",
                center=(26549.63, -41028.504, 7713.7764),
                radius=3000.0,
            ),
            Strongpoint(
                name="Eastern Slope",
                center=(29662.375, -3406.8445, 8725.453),
                radius=3000.0,
            ),
            Strongpoint(
                name="Trainwreck",
                center=(32129.537, 43600.098, 994.9547),
                radius=3000.0,
            ),
        ),
        (
            Strongpoint(
                name="Roer River Crossing",
                center=(64685.836, -33321.977, -2164.823),
                radius=3000.0,
            ),
            Strongpoint(
                name="Zerkall",
                center=(78823.555, -9569.677, -2095.891),
                radius=5000.0,
            ),
            Strongpoint(
                name="PaperMill",
                center=(69319.79, 39032.61, -2095.891),
                radius=3000.0,
            ),
        ),
    ),
)

SECTORS_HILL400_SMALL = Sector.skirmish_layout(
    orientation=Orientation.HORIZONTAL,
    strongpoint=Strongpoint(
        name="HILL 400",
        center=(0.0, 1015.0, -519.0),
        radius=8000.0,
    ),
)

SECTORS_HURTGENFOREST_LARGE = Sector.large_layout(
    orientation=Orientation.HORIZONTAL,
    strongpoints=(
        (
            Strongpoint(
                name="The Masbauch Approach",
                center=(-74423.0, -46733.0, 4336.0),
                radius=4625.0,
            ),
            Strongpoint(
                name="Reserve Station",
                center=(-78776.0, 2238.0, 3895.0),
                radius=4500.0,
            ),
            Strongpoint(
                name="Lumber Yard",
                center=(-77356.0, 36029.0, 4122.0),
                radius=4000.0,
            ),
        ),
        (
            Strongpoint(
                name="Wehebach Overlook",
                center=(-38278.0, -34416.0, 6683.0),
                radius=5031.25,
            ),
            Strongpoint(
                name="Kall Trail",
                center=(-35755.0, 2459.0, 4007.0),
                radius=6000.0,
            ),
            Strongpoint(
                name="The Ruin",
                center=(-42793.0, 26141.0, 3879.0),
                radius=4400.0,
            ),
        ),
        (
            Strongpoint(
                name="North Pass",
                center=(6540.0, -49329.0, 215.0),
                radius=4347.4674749999995,
            ),
            Strongpoint(
                name="The Scar",
                center=(-6935.0, 3328.0, 1327.0),
                radius=3015.2502,
            ),
            Strongpoint(
                name="The Siegfried Line",
                center=(-3711.0, 42305.0, 2603.0),
                radius=4500.0,
            ),
        ),
        (
            Strongpoint(
                name="Hill 15",
                center=(45628.0, -34330.0, 4504.0),
                radius=-4500.0,
            ),
            Strongpoint(
                name="Jacob's Barn",
                center=(37658.0, 8531.0, 6550.0),
                radius=3500.0,
            ),
            Strongpoint(
                name="Salient 42",
                center=(40632.0, 50244.0, 6895.0),
                radius=3250.0,
            ),
        ),
        (
            Strongpoint(
                name="Grosshau Approach",
                center=(73663.0, -38895.0, 5297.0),
                radius=3750.0,
            ),
            Strongpoint(
                name="Hürtgen Approach",
                center=(67776.0, 6558.0, 6600.0),
                radius=3500.0,
            ),
            Strongpoint(
                name="Logging Camp",
                center=(64477.0, 51502.0, 6495.0),
                radius=3750.0,
            ),
        ),
    ),
)

SECTORS_KHARKOV_LARGE = Sector.large_layout(
    orientation=Orientation.VERTICAL,
    strongpoints=(
        (
            Strongpoint(
                name="Marsh Town",
                center=(-36517.52, -70661.75, -2300.2422),
                radius=3750.0,
            ),
            Strongpoint(
                name="Soviet Vantage Point",
                center=(8032.91, -70714.63, 402.14062),
                radius=3750.0,
            ),
            Strongpoint(
                name="German Fuel Dump",
                center=(41168.312, -70231.15, 3192.4492),
                radius=3750.0,
            ),
        ),
        (
            Strongpoint(
                name="Bitter Spring",
                center=(-37433.285, -38891.406, -2293.7441),
                radius=8000.0,
            ),
            Strongpoint(
                name="Lumber Works",
                center=(7916.1367, -39814.156, 279.73047),
                radius=3750.0,
            ),
            Strongpoint(
                name="Windmill Hillside",
                center=(46877.23, -41370.87, 2556.1875),
                radius=3750.0,
            ),
        ),
        (
            Strongpoint(
                name="Water Mill",
                center=(-36761.05, -3563.8867, -2120.2441),
                radius=5000.0,
            ),
            Strongpoint(
                name="St Mary",
                center=(6074.9873, -633.23, 911.6289),
                radius=6000.0,
            ),
            Strongpoint(
                name="Distillery",
                center=(44449.215, -4542.487, 2724.1992),
                radius=3750.0,
            ),
        ),
        (
            Strongpoint(
                name="River Crossing",
                center=(-27116.355, 40003.023, -890.95703),
                radius=3750.0,
            ),
            Strongpoint(
                name="Belgorod Outskirts",
                center=(8105.9688, 38673.008, 221.38281),
                radius=9000.0,
            ),
            Strongpoint(
                name="Lumberyard",
                center=(46774.79, 37490.91, 2052.8438),
                radius=3750.0,
            ),
        ),
        (
            Strongpoint(
                name="Wehrmacht Overlook",
                center=(-37313.91, 72972.37, -661.0508),
                radius=3750.0,
            ),
            Strongpoint(
                name="Hay Storage",
                center=(4240.6523, 71736.38, -1926.957),
                radius=3750.0,
            ),
            Strongpoint(
                name="Overpass",
                center=(41180.39, 70416.95, -41.4375),
                radius=3750.0,
            ),
        ),
    ),
)

SECTORS_KURSK_LARGE = Sector.large_layout(
    orientation=Orientation.VERTICAL,
    strongpoints=(
        (
            Strongpoint(
                name="Artillery Position",
                center=(-35117.0, -68921.0, 9323.0),
                radius=6000.0,
            ),
            Strongpoint(
                name="Grushki",
                center=(7070.0, -68141.0, 7093.0),
                radius=4960.5308,
            ),
            Strongpoint(
                name="Grushki Flank",
                center=(47151.0, -67169.0, 5786.0),
                radius=4500.0,
            ),
        ),
        (
            Strongpoint(
                name="Panzer's End",
                center=(-35117.0, -31958.0, 8935.0),
                radius=6000.0,
            ),
            Strongpoint(
                name="Defence In Depth",
                center=(1604.0, -34906.0, 7647.0),
                radius=7022.2216,
            ),
            Strongpoint(
                name="Listening Post",
                center=(40413.0, -36000.0, 5889.0),
                radius=7673.426,
            ),
        ),
        (
            Strongpoint(
                name="The Windmills",
                center=(-26712.39, -4842.251, 9948.998),
                radius=6000.0,
            ),
            Strongpoint(
                name="Yamki",
                center=(9609.0, 3974.0, 8754.0),
                radius=6973.061,
            ),
            Strongpoint(
                name="Oleg's House",
                center=(39754.0, 7774.0, 6623.0),
                radius=4500.0,
            ),
        ),
        (
            Strongpoint(
                name="Rudno",
                center=(-27089.0, 40069.0, 9949.0),
                radius=6000.0,
            ),
            Strongpoint(
                name="Destroyed Battery",
                center=(-990.0, 39981.0, 10190.0),
                radius=4500.0,
            ),
            Strongpoint(
                name="The Muddy Churn",
                center=(41089.0, 42772.0, 7983.0),
                radius=4500.0,
            ),
        ),
        (
            Strongpoint(
                name="Road To Kursk",
                center=(-31287.0, 68120.0, 9949.0),
                radius=4500.0,
            ),
            Strongpoint(
                name="Ammo Dump",
                center=(-1729.0, 66294.0, 9632.0),
                radius=5446.865,
            ),
            Strongpoint(
                name="Eastern Position",
                center=(36100.0, 65758.0, 8227.0),
                radius=6000.0,
            ),
        ),
    ),
)

SECTORS_MORTAIN_LARGE = Sector.large_layout(
    orientation=Orientation.HORIZONTAL,
    strongpoints=(
        (
            Strongpoint(
                name="Hotel De La Poste",
                center=(-71664.03, -47217.445, -480.30115),
                radius=5000.0,
            ),
            Strongpoint(
                name="Forward Battery",
                center=(-67949.4, 6438.873, -978.4287),
                radius=6500.0,
            ),
            Strongpoint(
                name="Southern Approach",
                center=(-70344.09, 46402.33, -5391.659),
                radius=7500.0,
            ),
        ),
        (
            Strongpoint(
                name="Mortain Outskirts",
                center=(-49136.977, -39819.566, 1842.5009),
                radius=6000.0,
            ),
            Strongpoint(
                name="Forward Medical Aid Station",
                center=(-35275.574, -2194.1567, 2411.5898),
                radius=7000.0,
            ),
            Strongpoint(
                name="Mortain Approach",
                center=(-42775.5, 33050.027, -1580.7439),
                radius=7000.0,
            ),
        ),
        (
            Strongpoint(
                name="Hill 314",
                center=(-2425.1055, -38259.5, 5891.1714),
                radius=7000.0,
            ),
            Strongpoint(
                name="Petit Chappelle Saint Michel",
                center=(1725.2772, 5918.17, 5670.377),
                radius=5000.0,
            ),
            Strongpoint(
                name="U.S. Southern Roadblock",
                center=(-11254.05, 49076.438, -3064.87),
                radius=7000.0,
            ),
        ),
        (
            Strongpoint(
                name="Destroyed German Convoy",
                center=(35469.836, -42255.99, 6050.5566),
                radius=8000.0,
            ),
            Strongpoint(
                name="German Recon Camp",
                center=(40439.145, -2510.7285, 4398.437),
                radius=6000.0,
            ),
            Strongpoint(
                name="Le Hermitage Farm",
                center=(48018.547, 26619.574, 1071.811),
                radius=6000.0,
            ),
        ),
        (
            Strongpoint(
                name="Abandoned German Checkpoint",
                center=(68651.26, -40271.47, 6068.428),
                radius=6500.0,
            ),
            Strongpoint(
                name="German Defensive Camp",
                center=(68294.46, 1986.8845, 3941.4448),
                radius=7000.0,
            ),
            Strongpoint(
                name="Farm Of Bonovisin",
                center=(71327.67, 36841.695, 1896.9764),
                radius=7000.0,
            ),
        ),
    ),
)

SECTORS_MORTAIN_SMALL = Sector.skirmish_layout(
    orientation=Orientation.HORIZONTAL,
    strongpoint=Strongpoint(
        name="Petite Chappelle Saint Michel",
        center=(1000.0, 5500.0, -1000.0),
        radius=6000.0,
    ),
)

SECTORS_OMAHABEACH_LARGE = Sector.large_layout(
    orientation=Orientation.HORIZONTAL,
    strongpoints=(
        (
            Strongpoint(
                name="Beaumont Road",
                center=(-66508.0, -34528.0, 1461.6543),
                radius=5000.0,
            ),
            Strongpoint(
                name="Crossroads",
                center=(-63975.723, 2684.23, 1607.7931),
                radius=3713.8135,
            ),
            Strongpoint(
                name="Les Isles",
                center=(-65785.0, 33673.0, 1949.5156),
                radius=5000.0,
            ),
        ),
        (
            Strongpoint(
                name="Rear Battery",
                center=(-40364.508, -47019.88, 1471.3027),
                radius=5000.0,
            ),
            Strongpoint(
                name="Church Road",
                center=(-36692.0, -9308.0, 1471.3047),
                radius=5000.0,
            ),
            Strongpoint(
                name="The Orchards",
                center=(-44319.355, 27163.912, 1705.2588),
                radius=4000.0,
            ),
        ),
        (
            Strongpoint(
                name="West Vierville",
                center=(4665.0, -40540.0, 1262.0),
                radius=5000.0,
            ),
            Strongpoint(
                name="Vierville Sur Mer",
                center=(-2661.8896, -2895.0942, 889.7754),
                radius=5000.0,
            ),
            Strongpoint(
                name="Artillery Battery",
                center=(2342.277, 31510.633, 1730.7812),
                radius=5000.0,
            ),
        ),
        (
            Strongpoint(
                name="WN73",
                center=(54259.0, -44498.0, 164.8125),
                radius=5000.0,
            ),
            Strongpoint(
                name="WN71",
                center=(55132.387, -5791.973, 1015.4961),
                radius=3750.0,
            ),
            Strongpoint(
                name="WN70",
                center=(46516.0, 30340.0, 1368.2734),
                radius=5000.0,
            ),
        ),
        (
            Strongpoint(
                name="Dog Green",
                center=(67602.0, -31262.0, -3134.6387),
                radius=6250.0,
            ),
            Strongpoint(
                name="The Draw",
                center=(71322.0, -7432.0, -2748.504),
                radius=3750.0,
            ),
            Strongpoint(
                name="Dog White",
                center=(71817.0, 30284.0, -3019.504),
                radius=5000.0,
            ),
        ),
    ),
)

SECTORS_PURPLEHEARTLANE_LARGE = Sector.large_layout(
    orientation=Orientation.VERTICAL,
    strongpoints=(
        (
            Strongpoint(
                name="Bloody Bend",
                center=(-53699.133, -68803.984, 6831.375),
                radius=2750.0,
            ),
            Strongpoint(
                name="Dead Man's Corner",
                center=(740.8672, -65433.984, 6831.375),
                radius=4000.0,
            ),
            Strongpoint(
                name="Forward Battery",
                center=(33330.867, -66643.984, 6831.375),
                radius=4000.0,
            ),
        ),
        (
            Strongpoint(
                name="Jourdan Canal",
                center=(-41489.133, -38108.99, 6831.375),
                radius=2750.0,
            ),
            Strongpoint(
                name="Douve Bridge",
                center=(-1434.0474, -26826.268, 7010.586),
                radius=4250.0,
            ),
            Strongpoint(
                name="Douve River Battery",
                center=(33572.38, -36601.72, 6840.3447),
                radius=3500.0,
            ),
        ),
        (
            Strongpoint(
                name="Groult Pillbox",
                center=(-37607.4, -5672.991, 6730.6123),
                radius=5500.0,
            ),
            Strongpoint(
                name="Carentan Causeway",
                center=(787.74744, 1346.289, 6969.521),
                radius=3500.0,
            ),
            Strongpoint(
                name="Flak Position",
                center=(45592.906, -4116.6772, 6732.951),
                radius=-4750.0,
            ),
        ),
        (
            Strongpoint(
                name="Madeleine Farm",
                center=(-33264.676, 30204.594, 7391.552),
                radius=3250.0,
            ),
            Strongpoint(
                name="Madeleine Bridge",
                center=(1928.2188, 39878.098, 6973.26),
                radius=3000.0,
            ),
            Strongpoint(
                name="Aid Station",
                center=(47043.207, 32172.8, 6753.122),
                radius=3250.0,
            ),
        ),
        (
            Strongpoint(
                name="Ingouf Crossroads",
                center=(-36344.676, 66489.59, 7391.552),
                radius=3250.0,
            ),
            Strongpoint(
                name="Road To Carentan",
                center=(2953.2188, 63908.098, 6973.26),
                radius=3000.0,
            ),
            Strongpoint(
                name="Cabbage Patch",
                center=(46253.22, 62363.098, 6973.26),
                radius=2500.0,
            ),
        ),
    ),
)

SECTORS_PURPLEHEARTLANE_SMALL = Sector.skirmish_layout(
    orientation=Orientation.VERTICAL,
    strongpoint=Strongpoint(
        name="Carentan Causeway",
        center=(-960.0, 285.0, 46.0),
        radius=8000.0,
    ),
)

SECTORS_REMAGEN_LARGE = Sector.large_layout(
    orientation=Orientation.VERTICAL,
    strongpoints=(
        (
            Strongpoint(
                name="Alte Liebe Barsch",
                center=(-41114.0, -69583.0, 6515.0),
                radius=4000.0,
            ),
            Strongpoint(
                name="Bewaldet Kreuzung",
                center=(-891.0, -69550.0, 12708.0),
                radius=4000.0,
            ),
            Strongpoint(
                name="Dan Radart 512",
                center=(41625.0, -69063.0, 16150.0),
                radius=4000.0,
            ),
        ),
        (
            Strongpoint(
                name="Erpel",
                center=(-39275.0, -40853.0, 1774.0),
                radius=4000.0,
            ),
            Strongpoint(
                name="Erpeler Ley",
                center=(9697.0, -42679.0, 13960.0),
                radius=4000.0,
            ),
            Strongpoint(
                name="Kasbach Outlook",
                center=(38436.418, -41098.23, 9033.0),
                radius=4000.0,
            ),
        ),
        (
            Strongpoint(
                name="St Severin Chapel",
                center=(-39275.0, -12967.0, 766.0),
                radius=4000.0,
            ),
            Strongpoint(
                name="Ludendorff Bridge",
                center=(3032.2412, 7.0210953, 1261.005),
                radius=8000.0,
            ),
            Strongpoint(
                name="Bauernhof Am Rhein",
                center=(38817.02, 15613.944, 104.0),
                radius=4000.0,
            ),
        ),
        (
            Strongpoint(
                name="Remagen",
                center=(-35925.75, 39434.0, -27.363525),
                radius=4000.0,
            ),
            Strongpoint(
                name="Mobelfabrik",
                center=(-1000.0, 40824.0, -35.674072),
                radius=5000.0,
            ),
            Strongpoint(
                name="SchlieffenAusweg",
                center=(39053.0, 38264.0, 104.0),
                radius=4000.0,
            ),
        ),
        (
            Strongpoint(
                name="Waldburg",
                center=(-40954.977, 80279.71, -125.40869),
                radius=4000.0,
            ),
            Strongpoint(
                name="Muhlenweg",
                center=(3742.6152, 72094.91, -121.5036),
                radius=4000.0,
            ),
            Strongpoint(
                name="Hagelkreuz",
                center=(37607.746, 68933.32, 104.0),
                radius=4000.0,
            ),
        ),
    ),
)

SECTORS_STALINGRAD_LARGE = Sector.large_layout(
    orientation=Orientation.HORIZONTAL,
    strongpoints=(
        (
            Strongpoint(
                name="Mamayev Approach",
                center=(-69500.0, -47966.0, 5684.0),
                radius=7000.0,
            ),
            Strongpoint(
                name="Nail Factory",
                center=(-71016.0, 11068.0, 7295.0),
                radius=8000.0,
            ),
            Strongpoint(
                name="City Overlook",
                center=(-69346.0, 48417.0, 7445.0),
                radius=8000.0,
            ),
        ),
        (
            Strongpoint(
                name="Dolgiy Ravine",
                center=(-39681.0, -48845.0, 4095.0),
                radius=7500.0,
            ),
            Strongpoint(
                name="Yellow House",
                center=(-39693.438, -1.544678, 7515.0),
                radius=8000.0,
            ),
            Strongpoint(
                name="Komsomol HQ",
                center=(-39683.0, 39676.0, 7310.0),
                radius=8000.0,
            ),
        ),
        (
            Strongpoint(
                name="Railway Crossing",
                center=(7.0, -39673.0, 4505.0),
                radius=8000.0,
            ),
            Strongpoint(
                name="Carriage Depot",
                center=(-15.0, 13.0, 4961.0),
                radius=8500.0,
            ),
            Strongpoint(
                name="Train Station",
                center=(6.0, 39678.0, 4942.0),
                radius=8000.0,
            ),
        ),
        (
            Strongpoint(
                name="House Of The Workers",
                center=(36591.0, -40602.0, 4355.0),
                radius=8000.0,
            ),
            Strongpoint(
                name="Pavlov's House",
                center=(48586.0, 1452.0, 4035.0),
                radius=8000.0,
            ),
            Strongpoint(
                name="The Brewery",
                center=(39674.0, 41970.0, 4077.0),
                radius=8000.0,
            ),
        ),
        (
            Strongpoint(
                name="L-Shaped House",
                center=(68875.0, -35043.0, 4195.0),
                radius=8000.0,
            ),
            Strongpoint(
                name="Grudinin's Mill",
                center=(70063.0, -32.0, 3965.0),
                radius=8000.0,
            ),
            Strongpoint(
                name="Volga Banks",
                center=(70121.0, 43351.0, 3965.0),
                radius=8000.0,
            ),
        ),
    ),
)

SECTORS_STALINGRAD_SMALL = Sector.skirmish_layout(
    orientation=Orientation.HORIZONTAL,
    strongpoint=Strongpoint(
        name="Carriage Depot",
        center=(0.0, 0.0, -1388.4307),
        radius=6000.0,
    ),
)

SECTORS_STMARIEDUMONT_LARGE = Sector.large_layout(
    orientation=Orientation.VERTICAL,
    strongpoints=(
        (
            Strongpoint(
                name="Winters Landing",
                center=(-39503.258, -78343.03, 809.0),
                radius=5754.7485,
            ),
            Strongpoint(
                name="Le Grand Chemin",
                center=(-367.0, -76667.0, 809.0),
                radius=4500.0,
            ),
            Strongpoint(
                name="The Barn",
                center=(44896.0, -73822.0, -89.0),
                radius=5691.033,
            ),
        ),
        (
            Strongpoint(
                name="Brecourt Battery",
                center=(-39380.0, -39702.0, 809.0),
                radius=6078.6405,
            ),
            Strongpoint(
                name="Cattlesheds",
                center=(2961.319, -41557.402, 809.0),
                radius=5400.8775,
            ),
            Strongpoint(
                name="Rue De La Gare",
                center=(35565.562, -39370.594, 809.0),
                radius=5892.7635,
            ),
        ),
        (
            Strongpoint(
                name="The Dugout",
                center=(-37170.906, -151.18701, 460.30884),
                radius=5814.819,
            ),
            Strongpoint(
                name="AA Network",
                center=(1716.0, 2530.0, 422.64746),
                radius=6530.8635,
            ),
            Strongpoint(
                name="Pierre's Farm",
                center=(37508.1, 1336.6963, 438.40137),
                radius=5207.283,
            ),
        ),
        (
            Strongpoint(
                name="Hugo's Farm",
                center=(-38001.0, 38089.0, 809.0),
                radius=6046.29,
            ),
            Strongpoint(
                name="The Hamlet",
                center=(-2158.7668, 42649.312, 597.02246),
                radius=4500.0,
            ),
            Strongpoint(
                name="Ste Marie Du Mont",
                center=(47022.125, 50258.56, 1336.7599),
                radius=6000.0,
            ),
        ),
        (
            Strongpoint(
                name="The Corner",
                center=(-34620.76, 69152.766, 809.0),
                radius=5105.9565,
            ),
            Strongpoint(
                name="Hill 6",
                center=(142.14453, 76822.93, 467.88086),
                radius=4500.0,
            ),
            Strongpoint(
                name="The Fields",
                center=(39750.15, 78234.78, 1152.4478),
                radius=4500.0,
            ),
        ),
    ),
)

SECTORS_STMARIEDUMONT_SMALL = Sector.skirmish_layout(
    orientation=Orientation.VERTICAL,
    strongpoint=Strongpoint(
        name="CATTLESHEDS",
        center=(3670.0, -2995.0, 125.0),
        radius=10000.0,
    ),
)

SECTORS_STMEREEGLISE_LARGE = Sector.large_layout(
    orientation=Orientation.HORIZONTAL,
    strongpoints=(
        (
            Strongpoint(
                name="Flak Position",
                center=(-69311.0, -40772.0, -81.0),
                radius=4000.0,
            ),
            Strongpoint(
                name="Vaulaville",
                center=(-62223.0, -3146.0, -1163.0),
                radius=2507.38,
            ),
            Strongpoint(
                name="La Prairie",
                center=(-67517.0, 35037.0, 1050.0),
                radius=4000.0,
            ),
        ),
        (
            Strongpoint(
                name="Route Du Haras",
                center=(-40886.0, -37779.0, -688.63086),
                radius=4000.0,
            ),
            Strongpoint(
                name="Western Approach",
                center=(-32652.0, -14761.0, -400.0),
                radius=4000.0,
            ),
            Strongpoint(
                name="Rue De Gambosville",
                center=(-34553.0, 41733.0, -699.0),
                radius=4000.0,
            ),
        ),
        (
            Strongpoint(
                name="Hospice",
                center=(-1100.0, -46000.0, -400.0),
                radius=4000.0,
            ),
            Strongpoint(
                name="Ste Mere Eglise",
                center=(5949.0, -7436.0, -718.7539),
                radius=4741.136,
            ),
            Strongpoint(
                name="Checkpoint",
                center=(467.0, 32490.0, -1330.0),
                radius=4000.0,
            ),
        ),
        (
            Strongpoint(
                name="Artillery Battery",
                center=(39652.0, -34374.0, -400.0),
                radius=4000.0,
            ),
            Strongpoint(
                name="The Cemetery",
                center=(28858.0, 5593.0, -742.64844),
                radius=4000.0,
            ),
            Strongpoint(
                name="Maison Du Crique",
                center=(25884.0, 30530.0, -400.0),
                radius=4000.0,
            ),
        ),
        (
            Strongpoint(
                name="Les Vieux Vergers",
                center=(70168.0, -28861.0, -743.6719),
                radius=4000.0,
            ),
            Strongpoint(
                name="Cross Roads",
                center=(72279.0, 1393.0, -676.0),
                radius=4000.0,
            ),
            Strongpoint(
                name="Russeau De Ferme",
                center=(72138.0, 38912.0, -1125.0),
                radius=4000.0,
            ),
        ),
    ),
)

SECTORS_STMEREEGLISE_SMALL = Sector.skirmish_layout(
    orientation=Orientation.HORIZONTAL,
    strongpoint=Strongpoint(
        name="Sainte-Mère-Église",
        center=(165.0, -113.0, -76.0),
        radius=8000.0,
    ),
)

SECTORS_TOBRUK_LARGE = Sector.large_layout(
    orientation=Orientation.HORIZONTAL,
    strongpoints=(
        (
            Strongpoint(
                name="Guard Room",
                center=(-68855.0, -27530.0, -6107.8867),
                radius=7000.0,
            ),
            Strongpoint(
                name="Tank Graveyard",
                center=(-69405.0, -2075.0, -7493.6714),
                radius=8000.0,
            ),
            Strongpoint(
                name="Division Headquarters",
                center=(-69835.0, 45005.0, -8821.878),
                radius=7000.0,
            ),
        ),
        (
            Strongpoint(
                name="West Creek",
                center=(-40077.0, -44015.0, -5846.329),
                radius=8000.0,
            ),
            Strongpoint(
                name="Albergo Ristorante Moderno",
                center=(-29536.549, 2241.0, -7323.1304),
                radius=7000.0,
            ),
            Strongpoint(
                name="King Square",
                center=(-29485.152, 39744.316, -7213.4697),
                radius=8000.0,
            ),
        ),
        (
            Strongpoint(
                name="Desert Rat Caves",
                center=(31.221313, -39665.25, -5381.714),
                radius=8000.0,
            ),
            Strongpoint(
                name="Church Grounds",
                center=(59.92363, 11770.049, -7002.9688),
                radius=7000.0,
            ),
            Strongpoint(
                name="Admiralty House",
                center=(7919.4326, 48901.832, -7326.4785),
                radius=9000.0,
            ),
        ),
        (
            Strongpoint(
                name="Abandoned Ammo Cache",
                center=(39687.508, -39659.996, -4662.4526),
                radius=8000.0,
            ),
            Strongpoint(
                name="8th Army Medical Hospital",
                center=(40124.035, -2845.0, -7074.448),
                radius=9000.0,
            ),
            Strongpoint(
                name="Supply Drop",
                center=(39820.547, 43918.36, -7314.533),
                radius=9000.0,
            ),
        ),
        (
            Strongpoint(
                name="Road To Senussi Mine",
                center=(69522.8, -40790.0, -4568.7607),
                radius=7000.0,
            ),
            Strongpoint(
                name="Makeshift Aid Station",
                center=(69380.0, 25.0, -6849.7656),
                radius=9000.0,
            ),
            Strongpoint(
                name="Cargo Warehouses",
                center=(70047.33, 41171.96, -7616.6694),
                radius=8000.0,
            ),
        ),
    ),
)

SECTORS_TOBRUK_SMALL = Sector.skirmish_layout(
    orientation=Orientation.HORIZONTAL,
    strongpoint=Strongpoint(
        name="Church Grounds",
        center=(0.0, 7625.0, -844.0),
        radius=6000.0,
    ),
)

SECTORS_UTAHBEACH_LARGE = Sector.large_layout(
    orientation=Orientation.HORIZONTAL,
    strongpoints=(
        (
            Strongpoint(
                name="Mammut Radar",
                center=(-65158.0, -51522.0, -2401.0),
                radius=3000.0,
            ),
            Strongpoint(
                name="Flooded House",
                center=(-66464.0, -2944.0, -2388.0),
                radius=3000.0,
            ),
            Strongpoint(
                name="Sainte Marie Approach",
                center=(-64837.0, 52705.0, -2157.0),
                radius=3000.0,
            ),
        ),
        (
            Strongpoint(
                name="Sunken Bridge",
                center=(-30810.0, -47853.0, -2157.0),
                radius=3000.0,
            ),
            Strongpoint(
                name="La Grande Crique",
                center=(-28986.0, 508.0, -2388.0),
                radius=3294.095,
            ),
            Strongpoint(
                name="Drowned Fields",
                center=(-36400.0, 43928.0, -2157.0),
                radius=3000.0,
            ),
        ),
        (
            Strongpoint(
                name="WN4",
                center=(5359.0, -42267.0, -2283.0),
                radius=3742.354,
            ),
            Strongpoint(
                name="The Chapel",
                center=(10650.0, -7786.0, -2157.0),
                radius=3000.0,
            ),
            Strongpoint(
                name="WN7",
                center=(727.0, 50408.0, -2157.0),
                radius=5000.0,
            ),
        ),
        (
            Strongpoint(
                name="AA Battery",
                center=(35101.0, -43589.0, -2389.0),
                radius=3941.9941,
            ),
            Strongpoint(
                name="Hill 5",
                center=(36131.0, -1838.0, -2157.0),
                radius=3000.0,
            ),
            Strongpoint(
                name="WN5",
                center=(48763.0, 44080.0, -2247.0),
                radius=3194.091,
            ),
        ),
        (
            Strongpoint(
                name="Tare Green",
                center=(63845.0, -46581.0, -2284.0),
                radius=3000.0,
            ),
            Strongpoint(
                name="Red Roof House",
                center=(64923.67, 3144.1865, -2206.6382),
                radius=3250.0,
            ),
            Strongpoint(
                name="Uncle Red",
                center=(66675.0, 45162.0, -2157.0),
                radius=2823.963,
            ),
        ),
    ),
)



from datetime import datetime, timedelta
from enum import IntEnum, StrEnum
from typing import Annotated, Literal, NamedTuple

from pydantic import (
    AfterValidator,
    BaseModel,
    BeforeValidator,
    ConfigDict,
    Field,
    PlainValidator,
)
from pydantic.alias_generators import to_camel

from hllrcon.data import Faction, GameMode, Layer, Role, TimeOfDay

EmptyStringToNoneValidator = AfterValidator(lambda v: v or None)
SplitStringValidator = BeforeValidator(lambda x: str(x).split(",") if x else [])


class Response(BaseModel):
    model_config = ConfigDict(
        alias_generator=to_camel,
        validate_by_name=True,
    )


# TODO: Add console platforms
class PlayerPlatform(StrEnum):
    STEAM = "steam"
    """Steam"""

    EPIC = "epic"
    """Epic Games Store"""

    XBL = "xbl"
    """Xbox Game Pass"""

    PSN = "psn"
    """PlayStation Network"""

    PS5 = "ps5"
    """PlayStation 5"""

    XSX = "xsx"
    """Xbox Series X|S"""


# TODO: Add console platforms
class SupportedPlatform(StrEnum):
    STEAM = "Steam"
    PC_XBOX = "WinGDK"
    EPIC = "eos"


class PlayerFactionId(IntEnum):
    GER = 0
    US = 1
    SOV = 2
    CW = 3
    DAK = 4
    B8A = 5
    UNASSIGNED = 6


class PlayerRoleId(IntEnum):
    RIFLEMAN = 0
    ASSAULT = 1
    AUTOMATIC_RIFLEMAN = 2
    MEDIC = 3
    SPOTTER = 4
    SUPPORT = 5
    MACHINE_GUNNER = 6
    ANTI_TANK = 7
    ENGINEER = 8
    OFFICER = 9
    SNIPER = 10
    CREWMAN = 11
    TANK_COMMANDER = 12
    COMMANDER = 13


class ForceMode(IntEnum):
    IMMEDIATE = 0
    """Force the player to be switched immediately, killing them if currently alive."""

    # TODO: Verify behavior when player is already dead
    AFTER_DEATH = 1
    """Force the player to be switched upon death."""


class GetAdminLogResponseEntry(Response):
    timestamp: datetime
    message: str


class GetAdminLogResponse(Response):
    entries: list[GetAdminLogResponseEntry]
    """A list of log entries, oldest entries first."""


class GetCommandsResponseEntry(Response):
    id: str = Field(validation_alias="iD")
    friendly_name: str
    is_client_supported: bool


class GetCommandsResponse(Response):
    entries: list[GetCommandsResponseEntry]


class GetAdminGroupsResponse(Response):
    group_names: list[str]


class GetAdminUsersResponseEntry(Response):
    user_id: str
    group: str
    comment: str


class GetAdminUsersResponse(Response):
    admin_users: list[GetAdminUsersResponseEntry]


class GetBansResponseEntry(Response):
    user_id: str
    user_name: str
    time_of_banning: datetime
    duration_hours: int
    ban_reason: str
    admin_name: str


class GetBansResponse(Response):
    ban_list: list[GetBansResponseEntry]


class GetPlayerResponseScoreData(Response):
    combat: int = Field(validation_alias="cOMBAT")
    offense: int
    defense: int
    support: int


class GetPlayerResponseWorldPosition(NamedTuple):
    x: float
    """The east-west horizontal axis. Between -100000 and 100000."""

    y: float
    """The north-south horizontal axis. Between -100000 and 100000."""

    z: float
    """The vertical axis."""


class GetPlayerResponse(Response):
    name: str
    """The player's name"""

    clan_tag: Annotated[str | None, EmptyStringToNoneValidator]
    """The player's clan tag. Empty string if none."""

    id: str = Field(validation_alias="iD")
    """The player's ID"""

    platform: PlayerPlatform
    """The player's platform"""

    eos_id: str = Field(validation_alias="eosId")
    """The player's Epic Online Services ID"""

    level: int
    """The player's level"""

    faction_id: PlayerFactionId = Field(validation_alias="team")
    """The ID of the player's faction."""

    role_id: PlayerRoleId | int = Field(validation_alias="role")
    """The ID of the player's role."""

    platoon: Annotated[str | None, EmptyStringToNoneValidator]
    """The name of the player's squad."""

    loadout: str
    """The player's current loadout. Might not be accurate if not spawned in."""

    kills: int
    """The player's kills"""

    deaths: int
    """The player's deaths"""

    score_data: GetPlayerResponseScoreData
    """The player's score"""

    world_position: GetPlayerResponseWorldPosition
    """The player's position in centimeters"""

    @property
    def faction(self) -> Faction | None:
        return Faction.by_id(self.faction_id)

    @property
    def role(self) -> Role:
        return Role.by_id(self.role_id)


class GetPlayersResponse(Response):
    players: list[GetPlayerResponse]


class GetMapRotationResponseEntry(Response):
    name: str
    game_mode_name: str = Field(validation_alias="gameMode")
    time_of_day: TimeOfDay | str
    id: Annotated[str, PlainValidator(lambda x: x.rsplit("/", 1)[-1])] = Field(
        validation_alias="iD",
    )
    position: int

    @property
    def game_mode(self) -> GameMode:
        if self.game_mode_name.startswith("Control Skirmish"):
            return GameMode.SKIRMISH
        if self.game_mode_name.endswith("Offensive"):
            return GameMode.OFFENSIVE

        return GameMode.by_id(self.game_mode_name)

    def find_layer(self, *, strict: bool = True) -> Layer:
        """Attempt to find the layer associated with this map rotation entry.

        Parameters
        ----------
        strict : bool, optional
            Whether to raise an exception if no such layer is known. If set to `False`,
            will attempt to generate a fallback value based on the ID. By default
            `True`.

        Returns
        -------
        Layer
            The layer associated with this map rotation entry.

        Raises
        ------
        ValueError
            No layer information is known about this entry.

        """
        return Layer.by_id(self.id, strict=strict)


class GetMapRotationResponse(Response):
    maps: list[GetMapRotationResponseEntry] = Field(validation_alias="mAPS")


class GetServerSessionResponse(Response):
    server_name: str
    map_name: str
    game_mode_id: str = Field(validation_alias="gameMode")
    remaining_match_time: Annotated[
        timedelta,
        PlainValidator(
            lambda x: x if isinstance(x, timedelta) else timedelta(seconds=int(x)),
        ),
    ]
    match_time: int
    allied_score: int
    axis_score: int
    player_count: int
    allied_player_count: int
    axis_player_count: int
    max_player_count: int
    queue_count: int
    max_queue_count: int
    vip_queue_count: int
    max_vip_queue_count: int

    @property
    def game_mode(self) -> GameMode:
        return GameMode.by_id(self.game_mode_id)


class GetServerConfigResponse(Response):
    server_name: str
    build_number: str
    build_revision: str
    supported_platforms: list[SupportedPlatform]
    password_protected: bool


class GetBannedWordsResponse(Response):
    banned_words: list[str]


class GetVipsResponse(Response):
    vips: list[str] = Field(validation_alias="vipPlayerIds")


class GetCommandDetailsResponseParameter(Response):
    type: Literal["Combo", "Text", "Number"]
    """The type of parameter"""

    name: str
    """The user-friendly name of the parameter"""

    id: str = Field(validation_alias="iD")
    """The name of the parameter"""

    display_member: Annotated[list[str], SplitStringValidator]
    """A list of user-friendly values for this parameter. Empty if `type` is not
    \"Combo\""""

    value_member: Annotated[list[str], SplitStringValidator]
    """A list of values for this parameter. An empty list if `type` is not \"Combo\""""


class GetCommandDetailsResponse(Response):
    name: str
    """Name of the command"""

    text: str
    """User-friendly name of the command"""

    description: str
    """Description of the command"""

    dialogue_parameters: list[GetCommandDetailsResponseParameter]
    """A list of parameters for this command"""



import json
from enum import IntEnum
from typing import Any, Self

from hllrcon.exceptions import HLLCommandError


class RconResponseStatus(IntEnum):
    """Enumeration of RCON response status codes."""

    OK = 200
    """The request was successful."""

    BAD_REQUEST = 400
    """The request was invalid."""

    UNAUTHORIZED = 401
    """Insufficient or invalid authorization."""

    INTERNAL_ERROR = 500
    """An internal server error occurred."""


class RconResponse:
    """Represents a RCON response."""

    def __init__(
        self,
        request_id: int,
        command: str,
        version: int,
        status_code: RconResponseStatus,
        status_message: str,
        content_body: str,
    ) -> None:
        """Initializes a new RCON response.

        Parameters
        ----------
        request_id : int
            The ID of the request this response corresponds to.
        command : str
            The command that was executed.
        version : int
            The version of the command.
        status_code : RconResponseStatus
            The status code of the response.
        status_message : str
            A message describing the status of the response.
        content_body : str
            The body of the response, potentially JSON-deserializable.

        """
        self.request_id = request_id
        self.name = command
        self.version = version
        self.status_code = status_code
        self.status_message = status_message
        self.content_body = content_body

    @property
    def content_dict(self) -> dict[str, Any]:
        """JSON-deserialize the content body of the response.

        Raises
        ------
        json.JSONDecodeError
            The content body could not be deserialized.
        TypeError
            The deserialized content is not a dictionary.

        Returns
        -------
        dict[str, Any]
            The deserialized content body as a dictionary.

        """
        parsed_content = json.loads(self.content_body)
        if not isinstance(parsed_content, dict):
            msg = f"Expected JSON content to be a dict, got {type(parsed_content)}"
            raise TypeError(msg)
        return parsed_content

    def __str__(self) -> str:
        content: str | dict[str, Any]
        try:
            content = self.content_dict
        except (json.JSONDecodeError, TypeError):
            content = self.content_body

        return f"{self.status_code} {self.name} {content}"

    @classmethod
    def unpack(cls, request_id: int, body_encoded: bytes) -> Self:
        """Unpacks a RCON response from its bytes representation.

        Parameters
        ----------
        request_id : int
            The ID of the request this response corresponds to.
        body_encoded : bytes
            The encoded body of the response, which is expected to be a JSON string.

        Returns
        -------
        RconResponse
            The unpacked RCON response object.

        """
        body = json.loads(body_encoded)
        return cls(
            request_id=request_id,
            command=str(body["name"]),
            version=int(body["version"]),
            status_code=RconResponseStatus(int(body["statusCode"])),
            status_message=str(body["statusMessage"]),
            content_body=body["contentBody"],
        )

    def raise_for_status(self) -> None:
        """Raises an exception if the response status is not OK.

        Raises
        ------
        HLLCommandError
            The response status code is not `RconResponseStatus.OK`.

        """
        if self.status_code != RconResponseStatus.OK:
            raise HLLCommandError(self.status_code, self.status_message)



from collections.abc import Callable, Hashable
from typing import TYPE_CHECKING, Any, ClassVar, Generic, Never, Self, cast

from pydantic import BaseModel, ConfigDict, PrivateAttr
from typing_extensions import TypeVar

T = TypeVar("T")
H = TypeVar("H", bound=Hashable)
R = TypeVar("R", default=Never)


class class_cached_property(Generic[T]):  # noqa: N801
    def __init__(self, func: Callable[[type], T]) -> None:
        if TYPE_CHECKING or not isinstance(func, classmethod):  # pragma: no cover
            self.func = func
        else:
            self.func = func.__func__
        self.resolved: bool = False
        self.value: T | None = None

    def __get__(self, instance: T | None, owner: type) -> T:
        if not self.resolved:
            self.resolve(owner)
        return self.value  # type: ignore[return-value]

    def resolve(self, cls: type) -> None:
        self.value = self.func(cls)
        self.resolved = True


class IndexedBaseModel(BaseModel, Generic[H, R]):
    model_config = ConfigDict(
        ignored_types=(class_cached_property,),
        frozen=True,
    )
    _lookup_map: ClassVar[dict[Any, "IndexedBaseModel[Any, Any]"]]

    if TYPE_CHECKING:
        id: H = PrivateAttr()

    @classmethod
    def __pydantic_init_subclass__(cls, **kwargs: Any) -> None:  # noqa: ANN401
        if (
            cls.model_fields
            and "id" not in cls.model_fields
            and "id" not in cls.model_computed_fields
            and "id" not in cls.__dict__
        ):
            msg = f"{cls.__name__} must define an 'id' field."
            raise TypeError(msg)

        cls._lookup_map = {}

        for value in cls.__dict__.values():
            if isinstance(value, class_cached_property):
                value.resolve(cls)

    @classmethod
    def _lookup_register(cls, id_: H, instance: Self) -> None:
        if id_ in cls._lookup_map:
            msg = f"{cls.__name__} with ID {id_} already exists."
            raise ValueError(msg)

        cls._lookup_map[id_] = instance

    @classmethod
    def _lookup_fallback(cls, id_: H) -> Self | R:
        msg = f"{cls.__name__} with ID {id_} not found."
        raise ValueError(msg)

    def model_post_init(self, context: Any) -> None:  # noqa: ANN401, ARG002
        self._lookup_register(self.id, self)

    @classmethod
    def by_id(cls, id_: H) -> Self | R:
        """Look up a model by its identifier.

        Parameters
        ----------
        id_ : H
            The identifier of the model to look up.

        Returns
        -------
        Self | R
            The model instance with the given identifier.

        Raises
        ------
        ValueError
            If no model with the given identifier exists.

        """
        if instance := cls._lookup_map.get(id_):
            return cast("Self", instance)

        return cls._lookup_fallback(id_)

    @classmethod
    def all(cls) -> list[Self]:
        """Return all known instances of the model.

        Returns
        -------
        list[Self]
            A list of all model instances.

        """
        return list(cls._lookup_map.values())  # type: ignore[return-value]


class CaseInsensitiveIndexedBaseModel(IndexedBaseModel[str, R]):
    id: str

    @classmethod
    def by_id(cls, id_: str) -> Self | R:
        return super().by_id(id_.lower())

    @classmethod
    def _lookup_register(cls, id_: str, instance: Self) -> None:
        return super()._lookup_register(id_.lower(), instance)



# ruff: noqa: N802, D400, D415

from enum import StrEnum

from pydantic import BaseModel, Field

from hllrcon.data._utils import IndexedBaseModel, class_cached_property
from hllrcon.data.factions import Faction
from hllrcon.data.roles import Role, RoleType
from hllrcon.data.weapons import Weapon

_TANK_CREW_ROLES = {role for role in Role.all() if role.type == RoleType.ARMOR}


class VehicleType(StrEnum):
    HEAVY_TANK = "Heavy Tank"
    MEDIUM_TANK = "Medium Tank"
    LIGHT_TANK = "Light Tank"
    RECON_VEHICLE = "Recon Vehicle"
    HALF_TRACK = "Half-Track"
    TRANSPORT_TRUCK = "Transport Truck"
    SUPPLY_TRUCK = "Supply Truck"
    JEEP = "Jeep"
    ARTILLERY = "Artillery"
    ANTI_TANK_GUN = "Anti-Tank Gun"


class VehicleSeatType(StrEnum):
    DRIVER = "Driver"
    GUNNER = "Gunner"
    SPOTTER = "Spotter"
    LOADER = "Loader"
    PASSENGER = "Passenger"


class VehicleSeat(BaseModel, frozen=True):
    index: int
    type: VehicleSeatType
    weapons: list[Weapon] = Field(default_factory=list)
    requires_roles: set[Role] | None = None
    exposed: bool


class Vehicle(IndexedBaseModel[str]):
    id: str
    name: str
    factions: set[Faction] = Field(min_length=1)
    type: VehicleType
    seats: list[VehicleSeat]

    @class_cached_property
    @classmethod
    def M1_57MM(cls) -> "Vehicle":
        """*M1 57mm*"""
        return cls(
            id="M1 57mm",
            name="M1 57mm",
            type=VehicleType.ANTI_TANK_GUN,
            factions={Faction.US},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.GUNNER,
                    weapons=[Weapon.V_57MM_CANNON__M1_57],
                    exposed=True,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.LOADER,
                    exposed=True,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def M114(cls) -> "Vehicle":
        """*M114*"""
        return cls(
            id="M114",
            name="M114 Howitzer",
            type=VehicleType.ARTILLERY,
            factions={Faction.US},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.GUNNER,
                    weapons=[Weapon.V_57MM_CANNON__M1_57],
                    exposed=True,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.LOADER,
                    exposed=True,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def M8_GREYHOUND(cls) -> "Vehicle":
        """*M8 Greyhound*"""
        return cls(
            id="M8 Greyhound",
            name="M8 Greyhound",
            type=VehicleType.RECON_VEHICLE,
            factions={Faction.US},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.DRIVER,
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.GUNNER,
                    weapons=[
                        Weapon.V_M6_37MM__M8_GREYHOUND,
                        Weapon.V_COAXIAL_M1919__M8_GREYHOUND,
                    ],
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=2,
                    type=VehicleSeatType.SPOTTER,
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=3,
                    type=VehicleSeatType.PASSENGER,
                    exposed=False,
                ),
                VehicleSeat(
                    index=4,
                    type=VehicleSeatType.PASSENGER,
                    exposed=False,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def STUART_M5A1(cls) -> "Vehicle":
        """*Stuart M5A1*"""
        return cls(
            id="Stuart M5A1",
            name="M5A1 Stuart",
            type=VehicleType.LIGHT_TANK,
            factions={Faction.US},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.DRIVER,
                    weapons=[
                        Weapon.V_HULL_M1919__STUART_M5A1,
                    ],
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.GUNNER,
                    weapons=[
                        Weapon.V_37MM_CANNON__STUART_M5A1,
                        Weapon.V_COAXIAL_M1919__STUART_M5A1,
                    ],
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=2,
                    type=VehicleSeatType.SPOTTER,
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def SHERMAN_M4A3_75W(cls) -> "Vehicle":
        """*Sherman M4A3(75)W*"""
        return cls(
            id="Sherman M4A3(75)W",
            name="M4A3(75)W Sherman",
            type=VehicleType.MEDIUM_TANK,
            factions={Faction.US},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.DRIVER,
                    weapons=[
                        Weapon.V_HULL_M1919__SHERMAN_M4A3_75W,
                    ],
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.GUNNER,
                    weapons=[
                        Weapon.V_HULL_M1919__SHERMAN_M4A3_75W,
                    ],
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=2,
                    type=VehicleSeatType.SPOTTER,
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def SHERMAN_M4A3E2(cls) -> "Vehicle":
        """*Sherman M4A3E2*"""
        return cls(
            id="Sherman M4A3E2",
            name="M4A3E2 Sherman",
            type=VehicleType.HEAVY_TANK,
            factions={Faction.US},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.DRIVER,
                    weapons=[
                        Weapon.V_HULL_M1919__SHERMAN_M4A3E2,
                    ],
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.GUNNER,
                    weapons=[
                        Weapon.V_75MM_M3_GUN__SHERMAN_M4A3E2,
                        Weapon.V_COAXIAL_M1919__SHERMAN_M4A3E2,
                    ],
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=2,
                    type=VehicleSeatType.SPOTTER,
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def SHERMAN_M4A3E2_76(cls) -> "Vehicle":
        """*Sherman M4A3E2(76)*"""
        return cls(
            id="Sherman M4A3E2(76)",
            name="M4A3E2(76) Sherman",
            type=VehicleType.HEAVY_TANK,
            factions={Faction.US},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.DRIVER,
                    weapons=[
                        Weapon.V_HULL_M1919__SHERMAN_M4A3E2_76,
                    ],
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.GUNNER,
                    weapons=[
                        Weapon.V_76MM_M1_GUN__SHERMAN_M4A3E2_76,
                        Weapon.V_COAXIAL_M1919__SHERMAN_M4A3E2_76,
                    ],
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=2,
                    type=VehicleSeatType.SPOTTER,
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def GMC_CCKW_353_SUPPLY(cls) -> "Vehicle":
        """*GMC CCKW 353 (Supply)*"""
        return cls(
            id="GMC CCKW 353 (Supply)",
            name="GMC CCKW 353",
            type=VehicleType.SUPPLY_TRUCK,
            factions={Faction.US},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.DRIVER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def GMC_CCKW_353_TRANSPORT(cls) -> "Vehicle":
        """*GMC CCKW 353 (Transport)*"""
        return cls(
            id="GMC CCKW 353 (Transport)",
            name="GMC CCKW 353",
            type=VehicleType.TRANSPORT_TRUCK,
            factions={Faction.US},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.DRIVER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=2,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=3,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=4,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=5,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=6,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=7,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=8,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=9,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=10,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=11,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def M3_HALF_TRACK(cls) -> "Vehicle":
        """*M3 Half-track*"""
        return cls(
            id="M3 Half-track",
            name="M3 Half-track",
            type=VehicleType.HALF_TRACK,
            factions={Faction.US},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.DRIVER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.GUNNER,
                    weapons=[
                        Weapon.V_M2_BROWNING__M3_HALF_TRACK,
                    ],
                    exposed=True,
                ),
                VehicleSeat(
                    index=2,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=3,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=4,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=5,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=6,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=7,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def JEEP_WILLYS(cls) -> "Vehicle":
        """*Jeep Willys*"""
        return cls(
            id="Jeep Willys",
            name="Willy's Jeep",
            type=VehicleType.JEEP,
            factions={Faction.US},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.DRIVER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=2,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=3,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def PAK_40(cls) -> "Vehicle":
        """*PAK 40*"""
        return cls(
            id="PAK 40",
            name="Pak 40",
            type=VehicleType.ANTI_TANK_GUN,
            factions={Faction.GER, Faction.DAK},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.GUNNER,
                    weapons=[Weapon.V_75MM_CANNON__PAK_40],
                    exposed=True,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.LOADER,
                    exposed=True,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def SFH_18(cls) -> "Vehicle":
        """*sFH 18*"""
        return cls(
            id="sFH 18",
            name="sFH 18",
            type=VehicleType.ARTILLERY,
            factions={Faction.GER, Faction.DAK},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.GUNNER,
                    weapons=[Weapon.V_150MM_HOWITZER__SFH_18],
                    exposed=True,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.LOADER,
                    exposed=True,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def SD_KFZ_234_PUMA(cls) -> "Vehicle":
        """*Sd.Kfz.234 Puma*"""
        return cls(
            id="Sd.Kfz.234 Puma",
            name="Sd.Kfz.234 Puma",
            type=VehicleType.RECON_VEHICLE,
            factions={Faction.GER, Faction.DAK},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.DRIVER,
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.GUNNER,
                    weapons=[
                        Weapon.V_50MM_KWK_91_1__SD_KFZ_234_PUMA,
                        Weapon.V_COAXIAL_MG34__SD_KFZ_234_PUMA,
                    ],
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=2,
                    type=VehicleSeatType.SPOTTER,
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=3,
                    type=VehicleSeatType.PASSENGER,
                    exposed=False,
                ),
                VehicleSeat(
                    index=4,
                    type=VehicleSeatType.PASSENGER,
                    exposed=False,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def SD_KFZ_121_LUCHS(cls) -> "Vehicle":
        """*Sd.Kfz.121 Luchs*"""
        return cls(
            id="Sd.Kfz.121 Luchs",
            name="Sd.Kfz.121 Luchs",
            type=VehicleType.LIGHT_TANK,
            factions={Faction.GER, Faction.DAK},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.DRIVER,
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.GUNNER,
                    weapons=[
                        Weapon.V_20MM_KWK_30__SD_KFZ_121_LUCHS,
                        Weapon.V_COAXIAL_MG34__SD_KFZ_121_LUCHS,
                    ],
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=2,
                    type=VehicleSeatType.SPOTTER,
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def SD_KFZ_161_PANZER_IV(cls) -> "Vehicle":
        """*Sd.Kfz.161 Panzer IV*"""
        return cls(
            id="Sd.Kfz.161 Panzer IV",
            name="Sd.Kfz.161 Panzer IV",
            type=VehicleType.MEDIUM_TANK,
            factions={Faction.GER, Faction.DAK},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.DRIVER,
                    weapons=[
                        Weapon.V_HULL_MG34__SD_KFZ_161_PANZER_IV,
                    ],
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.GUNNER,
                    weapons=[
                        Weapon.V_75MM_CANNON__SD_KFZ_161_PANZER_IV,
                        Weapon.V_COAXIAL_MG34__SD_KFZ_161_PANZER_IV,
                    ],
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=2,
                    type=VehicleSeatType.SPOTTER,
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def SD_KFZ_171_PANTHER(cls) -> "Vehicle":
        """*Sd.Kfz.171 Panther*"""
        return cls(
            id="Sd.Kfz.171 Panther",
            name="Sd.Kfz.171 Panther",
            type=VehicleType.HEAVY_TANK,
            factions={Faction.GER, Faction.DAK},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.DRIVER,
                    weapons=[
                        Weapon.V_HULL_MG34__SD_KFZ_171_PANTHER,
                    ],
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.GUNNER,
                    weapons=[
                        Weapon.V_75MM_CANNON__SD_KFZ_171_PANTHER,
                        Weapon.V_COAXIAL_MG34__SD_KFZ_171_PANTHER,
                    ],
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=2,
                    type=VehicleSeatType.SPOTTER,
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def SD_KFZ_181_TIGER_1(cls) -> "Vehicle":
        """*Sd.Kfz.181 Tiger 1*"""
        return cls(
            id="Sd.Kfz.181 Tiger 1",
            name="Sd.Kfz.181 Tiger 1",
            type=VehicleType.HEAVY_TANK,
            factions={Faction.GER, Faction.DAK},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.DRIVER,
                    weapons=[
                        Weapon.V_HULL_MG34__SD_KFZ_181_TIGER_1,
                    ],
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.GUNNER,
                    weapons=[
                        Weapon.V_88MM_KWK_36_L_56__SD_KFZ_181_TIGER_1,
                        Weapon.V_COAXIAL_MG34__SD_KFZ_181_TIGER_1,
                    ],
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=2,
                    type=VehicleSeatType.SPOTTER,
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def OPEL_BLITZ_SUPPLY(cls) -> "Vehicle":
        """*Opel Blitz (Supply)*"""
        return cls(
            id="Opel Blitz (Supply)",
            name="Opel Blitz",
            type=VehicleType.SUPPLY_TRUCK,
            factions={Faction.GER, Faction.DAK},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.DRIVER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def OPEL_BLITZ_TRANSPORT(cls) -> "Vehicle":
        """*Opel Blitz (Transport)*"""
        return cls(
            id="Opel Blitz (Transport)",
            name="Opel Blitz",
            type=VehicleType.TRANSPORT_TRUCK,
            factions={Faction.GER, Faction.DAK},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.DRIVER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=2,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=3,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=4,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=5,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=6,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=7,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=8,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=9,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=10,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=11,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def SD_KFZ_251_HALF_TRACK(cls) -> "Vehicle":
        """*Sd.Kfz 251 Half-track*"""
        return cls(
            id="Sd.Kfz 251 Half-track",
            name="Sd.Kfz 251 Half-track",
            type=VehicleType.HALF_TRACK,
            factions={Faction.GER, Faction.DAK},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.DRIVER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.GUNNER,
                    weapons=[
                        Weapon.V_M2_BROWNING__M3_HALF_TRACK,
                    ],
                    exposed=True,
                ),
                VehicleSeat(
                    index=2,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=3,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=4,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=5,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=6,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=7,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def KUBELWAGEN(cls) -> "Vehicle":
        """*Kubelwagen*"""
        return cls(
            id="Kubelwagen",
            name="Kubelwagen",
            type=VehicleType.JEEP,
            factions={Faction.GER, Faction.DAK},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.DRIVER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=2,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=3,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def ZIS_2(cls) -> "Vehicle":
        """*ZiS-2*"""
        return cls(
            id="ZiS-2",
            name="ZiS-2",
            type=VehicleType.ANTI_TANK_GUN,
            factions={Faction.SOV},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.GUNNER,
                    weapons=[Weapon.V_57MM_CANNON__ZIS_2],
                    exposed=True,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.LOADER,
                    exposed=True,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def M1938_M_30(cls) -> "Vehicle":
        """*M1938 (M-30)*"""
        return cls(
            id="M1938 (M-30)",
            name="M-30",
            type=VehicleType.ARTILLERY,
            factions={Faction.SOV},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.GUNNER,
                    weapons=[Weapon.V_122MM_HOWITZER__M1938_M_30],
                    exposed=True,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.LOADER,
                    exposed=True,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def BA_10(cls) -> "Vehicle":
        """*BA-10*"""
        return cls(
            id="BA-10",
            name="BA-10",
            type=VehicleType.RECON_VEHICLE,
            factions={Faction.SOV},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.DRIVER,
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.GUNNER,
                    weapons=[
                        Weapon.V_19_K_45MM__BA_10,
                        Weapon.V_COAXIAL_DT__BA_10,
                    ],
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=2,
                    type=VehicleSeatType.SPOTTER,
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=3,
                    type=VehicleSeatType.PASSENGER,
                    exposed=False,
                ),
                VehicleSeat(
                    index=4,
                    type=VehicleSeatType.PASSENGER,
                    exposed=False,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def T70(cls) -> "Vehicle":
        """*T70*"""
        return cls(
            id="T70",
            name="T70",
            type=VehicleType.LIGHT_TANK,
            factions={Faction.SOV},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.DRIVER,
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.GUNNER,
                    weapons=[
                        Weapon.V_45MM_M1937__T70,
                        Weapon.V_COAXIAL_DT__T70,
                    ],
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=2,
                    type=VehicleSeatType.SPOTTER,
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def T34_76(cls) -> "Vehicle":
        """*T34/76*"""
        return cls(
            id="T34/76",
            name="T34/76",
            type=VehicleType.MEDIUM_TANK,
            factions={Faction.SOV},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.DRIVER,
                    weapons=[
                        Weapon.V_HULL_DT__T34_76,
                    ],
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.GUNNER,
                    weapons=[
                        Weapon.V_76MM_ZIS_5__T34_76,
                        Weapon.V_COAXIAL_DT__T34_76,
                    ],
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=2,
                    type=VehicleSeatType.SPOTTER,
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def IS_1(cls) -> "Vehicle":
        """*IS-1*"""
        return cls(
            id="IS-1",
            name="IS-1",
            type=VehicleType.HEAVY_TANK,
            factions={Faction.SOV},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.DRIVER,
                    weapons=[
                        Weapon.V_HULL_DT__IS_1,
                    ],
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.GUNNER,
                    weapons=[
                        Weapon.V_D_5T_85MM__IS_1,
                        Weapon.V_COAXIAL_DT__IS_1,
                    ],
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=2,
                    type=VehicleSeatType.SPOTTER,
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def ZIS_5_SUPPLY(cls) -> "Vehicle":
        """*ZIS-5 (Supply)*"""
        return cls(
            id="ZIS-5 (Supply)",
            name="ZIS-5",
            type=VehicleType.SUPPLY_TRUCK,
            factions={Faction.SOV},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.DRIVER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def ZIS_5_TRANSPORT(cls) -> "Vehicle":
        """*ZIS-5 (Transport)*"""
        return cls(
            id="ZIS-5 (Transport)",
            name="ZIS-5",
            type=VehicleType.TRANSPORT_TRUCK,
            factions={Faction.SOV},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.DRIVER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=2,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=3,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=4,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=5,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=6,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=7,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=8,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=9,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=10,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=11,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def GAZ_67(cls) -> "Vehicle":
        """*GAZ-67*"""
        return cls(
            id="GAZ-67",
            name="GAZ-67",
            type=VehicleType.JEEP,
            factions={Faction.SOV},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.DRIVER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=2,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=3,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def QF_6_POUNDER(cls) -> "Vehicle":
        """*QF 6-Pounder*"""
        return cls(
            id="QF 6-Pounder",
            name="QF 6-Pounder",
            type=VehicleType.ANTI_TANK_GUN,
            factions={Faction.CW, Faction.B8A},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.GUNNER,
                    weapons=[Weapon.V_QF_6_POUNDER__QF_6_POUNDER],
                    exposed=True,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.LOADER,
                    exposed=True,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def QF_25_POUNDER(cls) -> "Vehicle":
        """*QF 25-Pounder*"""
        return cls(
            id="QF 25-Pounder",
            name="QF 25-Pounder",
            type=VehicleType.ARTILLERY,
            factions={Faction.CW, Faction.B8A},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.GUNNER,
                    weapons=[Weapon.V_QF_25_POUNDER__QF_25_POUNDER],
                    exposed=True,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.LOADER,
                    exposed=True,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def DAIMLER(cls) -> "Vehicle":
        """*Daimler*"""
        return cls(
            id="Daimler",
            name="Daimler",
            type=VehicleType.RECON_VEHICLE,
            factions={Faction.CW, Faction.B8A},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.DRIVER,
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.GUNNER,
                    weapons=[
                        Weapon.V_QF_2_POUNDER__DAIMLER,
                        Weapon.V_COAXIAL_BESA__DAIMLER,
                    ],
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=2,
                    type=VehicleSeatType.SPOTTER,
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=3,
                    type=VehicleSeatType.PASSENGER,
                    exposed=False,
                ),
                VehicleSeat(
                    index=4,
                    type=VehicleSeatType.PASSENGER,
                    exposed=False,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def TETRARCH(cls) -> "Vehicle":
        """*Tetrarch*"""
        return cls(
            id="Tetrarch",
            name="Tetrarch",
            type=VehicleType.LIGHT_TANK,
            factions={Faction.CW},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.DRIVER,
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.GUNNER,
                    weapons=[
                        Weapon.V_QF_2_POUNDER__TETRARCH,
                        Weapon.V_COAXIAL_BESA__TETRARCH,
                    ],
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=2,
                    type=VehicleSeatType.SPOTTER,
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def M3_STUART_HONEY(cls) -> "Vehicle":
        """*M3 Stuart Honey*"""
        return cls(
            id="M3 Stuart Honey",
            name="M3 Stuart Honey",
            type=VehicleType.LIGHT_TANK,
            factions={Faction.B8A},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.DRIVER,
                    weapons=[
                        Weapon.V_HULL_M1919__M3_STUART_HONEY,
                    ],
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.GUNNER,
                    weapons=[
                        Weapon.V_37MM_CANNON__M3_STUART_HONEY,
                        Weapon.V_COAXIAL_M1919__M3_STUART_HONEY,
                    ],
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=2,
                    type=VehicleSeatType.SPOTTER,
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def CROMWELL(cls) -> "Vehicle":
        """*Cromwell*"""
        return cls(
            id="Cromwell",
            name="Cromwell",
            type=VehicleType.MEDIUM_TANK,
            factions={Faction.CW},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.DRIVER,
                    weapons=[
                        Weapon.V_HULL_BESA__CROMWELL,
                    ],
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.GUNNER,
                    weapons=[
                        Weapon.V_OQF_75MM__CROMWELL,
                        Weapon.V_COAXIAL_BESA__CROMWELL,
                    ],
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=2,
                    type=VehicleSeatType.SPOTTER,
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def CRUSADER_MK_III(cls) -> "Vehicle":
        """*Crusader Mk.III*"""
        return cls(
            id="Crusader Mk.III",
            name="Crusader Mk III",
            type=VehicleType.MEDIUM_TANK,
            factions={Faction.B8A},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.DRIVER,
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.GUNNER,
                    weapons=[
                        Weapon.V_OQF_57MM__CRUSADER_MK_III,
                        Weapon.V_COAXIAL_BESA__CRUSADER_MK_III,
                    ],
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=2,
                    type=VehicleSeatType.SPOTTER,
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def FIREFLY(cls) -> "Vehicle":
        """*Firefly*"""
        return cls(
            id="Firefly",
            name="Sherman Firefly",
            type=VehicleType.HEAVY_TANK,
            factions={Faction.CW},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.DRIVER,
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.GUNNER,
                    weapons=[
                        Weapon.V_QF_17_POUNDER__FIREFLY,
                        Weapon.V_COAXIAL_M1919__FIREFLY,
                    ],
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=2,
                    type=VehicleSeatType.SPOTTER,
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def CHURCHILL_MK_III(cls) -> "Vehicle":
        """*Churchill Mk.III*"""
        return cls(
            id="Churchill Mk.III",
            name="Churchill Mk III",
            type=VehicleType.HEAVY_TANK,
            factions={Faction.B8A},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.DRIVER,
                    weapons=[
                        Weapon.V_HULL_BESA_7_92MM__CHURCHILL_MK_III,
                    ],
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.GUNNER,
                    weapons=[
                        Weapon.V_OQF_57MM__CHURCHILL_MK_III,
                        Weapon.V_COAXIAL_BESA_7_92MM__CHURCHILL_MK_III,
                    ],
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=2,
                    type=VehicleSeatType.SPOTTER,
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def CHURCHILL_MK_VII(cls) -> "Vehicle":
        """*Churchill Mk.VII*"""
        return cls(
            id="Churchill Mk.VII",
            name="Churchill Mk VII",
            type=VehicleType.HEAVY_TANK,
            factions={Faction.CW},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.DRIVER,
                    weapons=[
                        Weapon.V_HULL_BESA_7_92MM__CHURCHILL_MK_VII,
                    ],
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.GUNNER,
                    weapons=[
                        Weapon.V_OQF_75MM__CHURCHILL_MK_VII,
                        Weapon.V_COAXIAL_BESA_7_92MM__CHURCHILL_MK_VII,
                    ],
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
                VehicleSeat(
                    index=2,
                    type=VehicleSeatType.SPOTTER,
                    requires_roles=_TANK_CREW_ROLES,
                    exposed=False,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def BEDFORD_OYD_SUPPLY(cls) -> "Vehicle":
        """*Bedford OYD (Supply)*"""
        return cls(
            id="Bedford OYD (Supply)",
            name="Bedford OYD",
            type=VehicleType.SUPPLY_TRUCK,
            factions={Faction.CW, Faction.B8A},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.DRIVER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
            ],
        )

    @class_cached_property
    @classmethod
    def BEDFORD_OYD_TRANSPORT(cls) -> "Vehicle":
        """*Bedford OYD (Transport)*"""
        return cls(
            id="Bedford OYD (Transport)",
            name="Bedford OYD",
            type=VehicleType.TRANSPORT_TRUCK,
            factions={Faction.CW, Faction.B8A},
            seats=[
                VehicleSeat(
                    index=0,
                    type=VehicleSeatType.DRIVER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=1,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=2,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=3,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=4,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=5,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=6,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=7,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=8,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=9,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=10,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
                VehicleSeat(
                    index=11,
                    type=VehicleSeatType.PASSENGER,
                    exposed=True,
                ),
            ],
        )



import asyncio
import logging
import threading
from collections.abc import Generator
from concurrent.futures import Future
from contextlib import contextmanager
from typing import Any

from typing_extensions import override

from hllrcon.rcon import Rcon
from hllrcon.sync.commands import SyncRconCommands


class SyncRcon(SyncRconCommands):
    """A synchronous interface for connecting to an RCON server.

    This is a wrapper for the asynchronous `Rcon` class, which is being run in a
    separate thread with its own event loop.

    To execute commands concurrently, a new `execute_concurrently` method is provided,
    which returns a `concurrent.futures.Future` object.
    """

    def __init__(
        self,
        host: str,
        port: int,
        password: str,
        logger: logging.Logger | None = None,
        reconnect_after_failures: int = 3,
    ) -> None:
        """Initialize a new `Rcon` instance.

        Parameters
        ----------
        host : str
            The hostname or IP address of the RCON server.
        port : int
            The port of the RCON server.
        password : str
            The password for the RCON server.
        logger : logging.Logger | None, optional
            A logger instance for logging messages, by default None. If None,
            `logging.getLogger(__name__)` is used.
        reconnect_after_failures : int, optional
            After how many failed attempts to execute a command the active connection is
            disposed and a new connection is established on the next command execution.
            If the server responds to a request, the failure count is reset, even if the
            server returned an error. Set to 0 to disable, by default 3.

        """
        self._logger = logger
        self._rcon = Rcon(
            host,
            port,
            password,
            logger=logger,
            reconnect_after_failures=reconnect_after_failures,
        )
        self._loop: asyncio.AbstractEventLoop | None = None

    @property
    def host(self) -> str:
        return self._rcon.host

    @host.setter
    def host(self, value: str) -> None:
        self._rcon.host = value

    @property
    def port(self) -> int:
        return self._rcon.port

    @port.setter
    def port(self, value: int) -> None:
        self._rcon.port = value

    @property
    def password(self) -> str:
        return self._rcon.password

    @password.setter
    def password(self, value: str) -> None:
        self._rcon.password = value

    @property
    def logger(self) -> logging.Logger:
        return self._logger or logging.getLogger(__name__)

    @logger.setter
    def logger(self, value: logging.Logger | None) -> None:
        self._logger = value
        self._rcon.logger = value

    @property
    def reconnect_after_failures(self) -> int:
        return self._rcon.reconnect_after_failures

    @reconnect_after_failures.setter
    def reconnect_after_failures(self, value: int) -> None:
        self._rcon.reconnect_after_failures = value

    def is_connected(self) -> bool:
        """Check if the client is connected to the RCON server.

        Returns
        -------
        bool
            True if connected, False otherwise.

        """
        return (
            self._loop is not None
            and self._loop.is_running()
            and self._rcon.is_connected()
        )

    @contextmanager
    def connect(self) -> Generator[None]:
        """Establish a connection to the RCON server."""
        self.wait_until_connected()
        try:
            yield
        finally:
            self.disconnect()

    def wait_until_connected(self) -> None:
        """Wait until the client is connected to the RCON server.

        This might be useful to verify that a connection can be established before
        continuing with other operations.
        """
        if self._loop is None or not self._loop.is_running():
            event = threading.Event()

            def target() -> None:
                self._loop = asyncio.new_event_loop()
                self._loop.call_soon(event.set)
                self._loop.run_forever()

            threading.Thread(
                target=target,
                name=f"SyncRconThread{id(self)}",
                daemon=True,
            ).start()

            # Wait for the loop to have started
            if not event.wait(timeout=1):
                if self._loop is not None:
                    self._loop.stop()
                    self._loop = None
                msg = "Thread never signalled back"
                raise RuntimeError(msg) from None

        if self._loop is None:  # pragma: no cover
            msg = "Could not run event loop"
            raise RuntimeError(msg)

        asyncio.run_coroutine_threadsafe(
            self._rcon.wait_until_connected(),
            loop=self._loop,
        ).result()

    def disconnect(self) -> None:
        """Disconnect from the RCON server."""
        self._rcon.disconnect()
        if self._loop is not None:
            self._loop.stop()
        self._loop = None

    def execute_concurrently(
        self,
        command: str,
        version: int,
        body: str | dict[str, Any] = "",
    ) -> Future[str]:
        """Schedule the execution of a command on the RCON server.

        This method allows for concurrent execution of commands.

        Parameters
        ----------
        command : str
            The command to execute.
        version : int
            The version of the command to execute.
        body : str | dict[str, Any], optional
            The body of the command, by default an empty string.

        Returns
        -------
        concurrent.futures.Future[str]
            A Future representing the response from the server.

        """
        self.wait_until_connected()

        if self._loop is None:  # pragma: no cover
            msg = "Could not run event loop"
            raise RuntimeError(msg)

        return asyncio.run_coroutine_threadsafe(
            self._rcon.execute(command, version, body),
            loop=self._loop,
        )

    @override
    def execute(
        self,
        command: str,
        version: int,
        body: str | dict[str, Any] = "",
    ) -> str:
        return self.execute_concurrently(
            command=command,
            version=version,
            body=body,
        ).result()



from collections.abc import Generator
from typing import NamedTuple

import pytest
from hllrcon.data import (
    Faction,
    GameMode,
    GameModeScale,
    Grid,
    Layer,
    Loadout,
    LoadoutId,
    Map,
    Role,
    Strongpoint,
    Team,
    TimeOfDay,
    Vehicle,
    Weapon,
    Weather,
)
from hllrcon.data._utils import IndexedBaseModel, class_cached_property
from hllrcon.data.sectors import GridPositionalModel
from pydantic import BaseModel, ValidationError


@pytest.fixture(autouse=True)
def reset_lookup_maps() -> Generator[None]:
    layer_lookup_map = Layer._lookup_map.copy()
    map_lookup_map = Map._lookup_map.copy()

    yield

    Layer._lookup_map = layer_lookup_map
    Map._lookup_map = map_lookup_map


class TestDataUtils:
    def test_indexed_model(self) -> None:
        class MyModel(IndexedBaseModel[int]):
            id: int
            name: str

        foo = MyModel(id=1, name="Foo")
        bar = MyModel(id=2, name="Bar")

        with pytest.raises(ValueError, match="already exists"):
            MyModel(id=2, name="Baz")

        assert MyModel.by_id(1) == foo
        assert MyModel.by_id(2) == bar

        with pytest.raises(ValueError, match="not found"):
            MyModel.by_id(3)

        assert MyModel.all() == [foo, bar]

    def test_indexed_model_resolves_properties(self) -> None:
        class MyModel(IndexedBaseModel[int]):
            id: int

            @class_cached_property
            @classmethod
            def bar(cls) -> "MyModel":
                return cls(id=2)

        assert MyModel.bar.all() == [MyModel.bar]

    def test_resolve_class_cache_property(self) -> None:
        class MyModel(BaseModel, ignored_types=(class_cached_property,)):
            id: int

            @class_cached_property
            @classmethod
            def foo(cls) -> "MyModel":
                return cls(id=3)

        assert MyModel.foo is MyModel.foo
        assert MyModel.foo.id == 3

    def test_missing_id(self) -> None:
        with pytest.raises(TypeError, match="must define an 'id' field."):

            class MyModel(IndexedBaseModel[int]):
                name: str


class TestDataFactions:
    def test_faction_by_id(self) -> None:
        assert Faction.by_id(0) == Faction.GER
        assert Faction.by_id(1) == Faction.US
        assert Faction.by_id(2) == Faction.SOV
        assert Faction.by_id(3) == Faction.CW
        assert Faction.by_id(4) == Faction.DAK
        assert Faction.by_id(5) == Faction.B8A
        assert Faction.by_id(6) is None

        with pytest.raises(ValueError, match="not found"):
            Faction.by_id(7)

        assert None not in Faction.all()

    class FactionProperties(NamedTuple):
        faction: Faction
        is_allied: bool
        is_axis: bool

    @pytest.mark.parametrize(
        "properties",
        [
            FactionProperties(Faction.GER, False, True),
            FactionProperties(Faction.US, True, False),
            FactionProperties(Faction.SOV, True, False),
            FactionProperties(Faction.CW, True, False),
            FactionProperties(Faction.DAK, False, True),
            FactionProperties(Faction.B8A, True, False),
        ],
    )
    def test_faction_properties(self, properties: FactionProperties) -> None:
        assert properties.faction.is_allied is properties.is_allied
        assert properties.faction.is_axis is properties.is_axis


class TestDataGameModes:
    def test_game_mode_by_id(self) -> None:
        assert GameMode.by_id("warfare") == GameMode.WARFARE
        assert GameMode.by_id("offensive") == GameMode.OFFENSIVE
        assert GameMode.by_id("skirmish") == GameMode.SKIRMISH

        assert GameMode.by_id("wArFaRe") == GameMode.WARFARE

        with pytest.raises(ValueError, match="not found"):
            GameMode.by_id("invalid_mode")

    def test_game_mode_scale(self) -> None:
        large = GameMode(id="large", scale=GameModeScale.LARGE)
        small = GameMode(id="small", scale=GameModeScale.SMALL)

        assert large.is_large()
        assert not large.is_small()

        assert not small.is_large()
        assert small.is_small()


class TestDataLayers:
    def test_layer_by_id(self) -> None:
        assert Layer.by_id("DRL_S_1944_Day_P_Skirmish") == Layer.DRIEL_SKIRMISH_DAY
        assert Layer.by_id("drl_s_1944_day_p_skirmish") == Layer.DRIEL_SKIRMISH_DAY

        with pytest.raises(ValueError, match="not parse"):
            Layer.by_id("Not a layer", strict=False)

        with pytest.raises(ValueError, match="not found"):
            Layer.by_id("Not a layer", strict=True)

    def test_layer_str(self) -> None:
        layer = Layer.DRIEL_SKIRMISH_DAY
        assert str(layer) == layer.id

    def test_layer_repr(self) -> None:
        layer = Layer.DRIEL_SKIRMISH_DAY
        expected_repr = (
            f"Layer(id='DRL_S_1944_Day_P_Skirmish', map={layer.map!r}, "
            f"attackers={layer.attacking_team!r}, time_of_day={layer.time_of_day!r}, "
            f"weather={layer.weather!r})"
        )
        assert repr(layer) == expected_repr

    def test_layer_equality(self) -> None:
        layer = Layer.DRIEL_SKIRMISH_DAY
        assert layer == layer  # noqa: PLR0124
        assert layer == layer.id.lower()
        assert layer != Layer.DRIEL_SKIRMISH_DAWN
        assert layer != "Some other layer"
        assert layer != 12345

    def test_layer_hash(self) -> None:
        layer = Layer.DRIEL_SKIRMISH_DAY
        assert hash(layer) == hash(layer.id.lower())
        assert hash(layer) != hash(Layer.DRIEL_SKIRMISH_DAWN)

    @pytest.mark.parametrize(
        ("layer", "name"),
        [
            (Layer.DRIEL_SKIRMISH_DAY, "Driel Skirmish"),
            (Layer.DRIEL_OFFENSIVE_CW_DAY, "Driel Off. CW"),
            (Layer.MORTAIN_OFFENSIVE_GER_OVERCAST, "Mortain Off. GER (Overcast)"),
            (Layer.STMARIEDUMONT_SKIRMISH_RAIN, "St. Marie Du Mont Skirmish (Rain)"),
            (Layer.ELALAMEIN_WARFARE_DUSK, "El Alamein Warfare (Dusk)"),
            (
                Layer.ELSENBORNRIDGE_OFFENSIVE_US_DAWN,
                "Elsenborn Ridge Off. US (Dawn, Snow)",
            ),
        ],
    )
    def test_layer_pretty_name(self, layer: Layer, name: str) -> None:
        assert layer.pretty_name == name

    def test_layer_pretty_name_missing_attacking_team(self) -> None:
        layer = Layer(
            id="test_layer",
            map=Map.KHARKOV,
            game_mode=GameMode.OFFENSIVE,
            time_of_day=TimeOfDay.DAY,
            weather=Weather.OVERCAST,
            attacking_team=None,
            grid=Grid.large(),
            sectors=[],
        )
        assert layer.pretty_name == "Kharkov Off. (Overcast)"

    def test_layer_attackers(self) -> None:
        layer1 = Layer.KURSK_OFFENSIVE_SOV_DAY
        assert layer1.attacking_team == Team.ALLIES
        assert layer1.defending_team == Team.AXIS
        assert layer1.attacking_faction == Faction.SOV
        assert layer1.defending_faction == Faction.GER

        layer2 = Layer.KURSK_OFFENSIVE_GER_DAY
        assert layer2.attacking_team == Team.AXIS
        assert layer2.defending_team == Team.ALLIES
        assert layer2.attacking_faction == Faction.GER
        assert layer2.defending_faction == Faction.SOV

        layer3 = Layer.KURSK_WARFARE_DAY
        assert layer3.attacking_team is None
        assert layer3.defending_team is None
        assert layer3.attacking_faction is None
        assert layer3.defending_faction is None

    def test_layer_parse_id_small(self) -> None:
        layer = Layer.by_id("FOO_S_1942_Rain_P_Skirmish", strict=False)
        assert layer.map.id == "FOO"
        assert layer.map.tag == "FOO"
        assert layer.map.name == "Foo"
        assert layer.game_mode == GameMode.SKIRMISH
        assert layer.time_of_day == TimeOfDay.DAY
        assert layer.weather == Weather.RAIN
        assert layer.attacking_faction is None

    def test_layer_parse_id_large(self) -> None:
        layer = Layer.by_id("FOO_L_1942_OffensiveCW_Morning", strict=False)
        assert layer.map.id == "FOO"
        assert layer.map.tag == "FOO"
        assert layer.map.name == "Foo"
        assert layer.game_mode == GameMode.OFFENSIVE
        assert layer.time_of_day == TimeOfDay.DAWN
        assert layer.weather == Weather.CLEAR
        assert layer.attacking_faction == Faction.CW

    def test_layer_parse_id_legacy(self) -> None:
        layer = Layer.by_id("foo_warfare_night", strict=False)
        assert layer.map.id == "foo"
        assert layer.map.tag == "FOO"
        assert layer.map.name == "Foo"
        assert layer.game_mode == GameMode.WARFARE
        assert layer.time_of_day == TimeOfDay.NIGHT
        assert layer.weather == Weather.CLEAR
        assert layer.attacking_faction is None

    def test_layer_parse_id_legacy_offensive(self) -> None:
        layer = Layer.by_id("foo_offensive_ger", strict=False)
        assert layer.map.id == "foo"
        assert layer.map.tag == "FOO"
        assert layer.map.name == "Foo"
        assert layer.game_mode == GameMode.OFFENSIVE
        assert layer.time_of_day == TimeOfDay.DAY
        assert layer.weather == Weather.CLEAR
        assert layer.attacking_faction == Faction.GER

    def test_layer_parse_id_unknown_game_mode(self) -> None:
        with pytest.raises(ValueError, match="not parse"):
            Layer.by_id("foy_notamode", strict=False)

    def test_layer_can_parse_known_layers(self) -> None:
        all_layers = Layer.all()
        Layer._lookup_map.clear()

        for layer in all_layers:
            Map._lookup_map.clear()
            Layer._parse_id(layer.id)


class TestSectors:
    def test_grid_to_world(self) -> None:
        grid = Grid(
            scale=20000,
            offset=(1000, 0),
            size=((-5, -5), (4, 4)),
        )

        assert grid.grid_to_world_from((1, 2)) == (21000, 40000)
        assert grid.grid_to_world_to((1, 2)) == (41000, 60000)

        assert grid.grid_to_world_from((-1, -2)) == (-19000, -40000)
        assert grid.grid_to_world_to((-1, -2)) == (1000, -20000)

    def test_world_to_grid(self) -> None:
        grid = Grid(
            scale=20000,
            offset=(1000, 0),
            size=((-5, -5), (4, 4)),
        )

        assert grid.world_to_grid((0, 0)) == (-1, 0)
        assert grid.world_to_grid((20000, 50000)) == (0, 2)
        assert grid.world_to_grid((-20000, -50000)) == (-2, -3)

    def test_grid_area(self) -> None:
        assert Layer.TOBRUK_WARFARE_DAY.sectors[0].area == (
            (-100000, -60000),
            (-60000, 60000),
        )

        assert Layer.PURPLEHEARTLANE_SKIRMISH_RAIN.sectors[0].area == (
            (-55705.6, -69632),
            (55705.6, -13926.4),
        )

    def test_validate_grid_coords_order(self) -> None:
        GridPositionalModel(grid_from=(0, 0), grid_to=(1, 1))
        GridPositionalModel(grid_from=(0, 0), grid_to=(0, 0))
        GridPositionalModel(grid_from=(-2, 0), grid_to=(-1, 3))

        with pytest.raises(
            ValidationError,
            match=r"grid_from must be smaller than grid_to",
        ):
            GridPositionalModel(grid_from=(1, 0), grid_to=(0, 1))

        with pytest.raises(
            ValidationError,
            match=r"grid_from must be smaller than grid_to",
        ):
            GridPositionalModel(grid_from=(0, 1), grid_to=(1, 0))

        with pytest.raises(
            ValidationError,
            match=r"grid_from must be smaller than grid_to",
        ):
            GridPositionalModel(grid_from=(1, 1), grid_to=(0, 0))

    def test_strongpoint_is_inside(self) -> None:
        sp = Strongpoint(
            name="Foo",
            center=(10, 0, 0),
            radius=10,
        )

        assert sp.is_inside((10, 0, 0))
        assert sp.is_inside((20, 0, 0))
        assert sp.is_inside((10, 10, 0))
        assert not sp.is_inside((-1, 0, 0))
        assert not sp.is_inside((10, 10, 10))

    def test_capture_zone_is_inside(self) -> None:
        zone = Layer.STMEREEGLISE_WARFARE_DAY.sectors[2].capture_zones[2]

        assert zone.is_inside((0, 40000))
        assert zone.is_inside((-19000, 21000))
        assert zone.is_inside((19000, 21000))
        assert zone.is_inside((19000, 59000))
        assert zone.is_inside((-19000, 59000))

        assert not zone.is_inside((0, 0))
        assert not zone.is_inside((-21000, 40000))

    def test_sector_is_inside(self) -> None:
        sector = Layer.KHARKOV_WARFARE_DAY.sectors[2]

        assert sector.is_inside((0, 0))
        assert sector.is_inside((0, 19000))
        assert not sector.is_inside((0, 21000))
        assert sector.is_inside((59000, 19000))
        assert not sector.is_inside((61000, 19000))


class TestDataMaps:
    def test_map_by_id(self) -> None:
        assert Map.by_id("kharkov") == Map.KHARKOV
        assert Map.by_id("driel") == Map.DRIEL
        assert Map.by_id("elalamein") == Map.EL_ALAMEIN
        assert Map.by_id("mortain") == Map.MORTAIN
        assert Map.by_id("elsenbornridge") == Map.ELSENBORN_RIDGE

        assert Map.by_id("kHaRkOv") == Map.KHARKOV

        with pytest.raises(ValueError, match="not found"):
            Map.by_id("invalid_map")

    def test_map_str(self) -> None:
        map_ = Map.KHARKOV
        assert str(map_) == map_.id

    def test_map_repr(self) -> None:
        map_ = Map.KHARKOV
        expected_repr = (
            f"Map(id='kharkov', name='Kharkov', tag='KHA', "
            f"pretty_name='Kharkov', short_name='Kharkov', "
            f"allies={map_.allies!r}, axis={map_.axis!r}, "
            f"orientation={map_.orientation!r}, "
            f"mirror_factions={map_.mirror_factions!r})"
        )
        assert repr(map_) == expected_repr

    def test_map_equality(self) -> None:
        map_ = Map.KHARKOV
        assert map_ == map_  # noqa: PLR0124
        assert map_ == map_.id.lower()
        assert map_ != Map.DRIEL
        assert map_ != "Some other map"
        assert map_ != 12345

    def test_map_hash(self) -> None:
        map_ = Map.KHARKOV
        assert hash(map_) == hash(map_.id.lower())
        assert hash(map_) != hash(Map.DRIEL)


class TestDataTeams:
    def test_team_by_id(self) -> None:
        assert Team.by_id(1) == Team.ALLIES
        assert Team.by_id(2) == Team.AXIS

        with pytest.raises(ValueError, match="not found"):
            Team.by_id(3)


class TestDataRoles:
    def test_role_by_id(self) -> None:
        assert Role.by_id(0) == Role.RIFLEMAN
        assert Role.by_id(1) == Role.ASSAULT
        assert Role.by_id(2) == Role.AUTOMATIC_RIFLEMAN
        assert Role.by_id(3) == Role.MEDIC
        assert Role.by_id(4) == Role.SPOTTER
        assert Role.by_id(5) == Role.SUPPORT
        assert Role.by_id(6) == Role.MACHINE_GUNNER
        assert Role.by_id(7) == Role.ANTI_TANK
        assert Role.by_id(8) == Role.ENGINEER
        assert Role.by_id(9) == Role.OFFICER
        assert Role.by_id(10) == Role.SNIPER
        assert Role.by_id(11) == Role.CREWMAN
        assert Role.by_id(12) == Role.TANK_COMMANDER
        assert Role.by_id(13) == Role.COMMANDER

        with pytest.raises(ValueError, match="not found"):
            Team.by_id(14)

    class RoleProperties(NamedTuple):
        role: Role
        is_infantry: bool
        is_tanker: bool
        is_recon: bool
        is_squad_leader: bool

    @pytest.mark.parametrize(
        "properties",
        [
            RoleProperties(Role.RIFLEMAN, True, False, False, False),
            RoleProperties(Role.ASSAULT, True, False, False, False),
            RoleProperties(Role.AUTOMATIC_RIFLEMAN, True, False, False, False),
            RoleProperties(Role.MEDIC, True, False, False, False),
            RoleProperties(Role.SPOTTER, False, False, True, True),
            RoleProperties(Role.SUPPORT, True, False, False, False),
            RoleProperties(Role.MACHINE_GUNNER, True, False, False, False),
            RoleProperties(Role.ANTI_TANK, True, False, False, False),
            RoleProperties(Role.ENGINEER, True, False, False, False),
            RoleProperties(Role.OFFICER, True, False, False, True),
            RoleProperties(Role.SNIPER, False, False, True, False),
            RoleProperties(Role.CREWMAN, False, True, False, False),
            RoleProperties(Role.TANK_COMMANDER, False, True, False, True),
            RoleProperties(Role.COMMANDER, False, False, False, True),
        ],
    )
    def test_role_properties(self, properties: RoleProperties) -> None:
        assert properties.role.is_infantry is properties.is_infantry
        assert properties.role.is_tanker is properties.is_tanker
        assert properties.role.is_recon is properties.is_recon
        assert properties.role.is_squad_leader is properties.is_squad_leader


class TestDataWeapons:
    def test_weapon_by_id(self) -> None:
        assert Weapon.by_id("M1 GARAND") == Weapon.M1_GARAND
        assert Weapon.by_id("MP40") == Weapon.MP40
        assert (
            Weapon.by_id("COAXIAL M1919 [Sherman M4A3E2]")
            == Weapon.V_COAXIAL_M1919__SHERMAN_M4A3E2
        )

        with pytest.raises(ValueError, match="not found"):
            Weapon.by_id("invalid_weapon")

        with pytest.raises(ValueError, match="not found"):
            Weapon.by_id("m1 garand")

    def test_weapon_resolve_vehicle(self) -> None:
        for weapon in Weapon.all():
            weapon.vehicle  # noqa: B018


class TestDataVehicles:
    def test_vehicle_by_id(self) -> None:
        assert Vehicle.by_id("Sherman M4A3E2") == Vehicle.SHERMAN_M4A3E2
        assert Vehicle.by_id("sFH 18") == Vehicle.SFH_18

        with pytest.raises(ValueError, match="not found"):
            Vehicle.by_id("invalid_vehicle")

        with pytest.raises(ValueError, match="not found"):
            Vehicle.by_id("sherman m4a3e2")


class TestDataLoadouts:
    def test_loadout_by_id(self) -> None:
        assert (
            Loadout.by_id(LoadoutId(Faction.US.id, Role.OFFICER.id, "NCO"))
            == Loadout.US_OFFICER_NCO
        )
        assert (
            Loadout.by_id(LoadoutId(Faction.CW.id, Role.RIFLEMAN.id, "Standard Issue"))
            == Loadout.CW_RIFLEMAN_STANDARD_ISSUE
        )

        assert (
            Loadout.by_id(LoadoutId(Faction.CW.id, Role.RIFLEMAN.id, "sTaNdArD iSsUe"))
            == Loadout.CW_RIFLEMAN_STANDARD_ISSUE
        )

        with pytest.raises(ValueError, match="not found"):
            Loadout.by_id(LoadoutId(Faction.US.id, Role.RIFLEMAN.id, "invalid_loadout"))
        with pytest.raises(ValueError, match="not found"):
            Loadout.by_id(LoadoutId(Faction.US.id, 69, "Standard Issue"))

    def test_loadout_by_name(self) -> None:
        assert (
            Loadout.by_name(Faction.US, Role.OFFICER, "NCO") == Loadout.US_OFFICER_NCO
        )
        assert (
            Loadout.by_name(Faction.CW, Role.RIFLEMAN, "Standard Issue")
            == Loadout.CW_RIFLEMAN_STANDARD_ISSUE
        )

        assert (
            Loadout.by_name(Faction.CW, Role.RIFLEMAN, "sTaNdArD iSsUe")
            == Loadout.CW_RIFLEMAN_STANDARD_ISSUE
        )

        with pytest.raises(ValueError, match="not found"):
            Loadout.by_name(Faction.US, Role.RIFLEMAN, "invalid_loadout")

    def test_loadout_equipment_weapon(self) -> None:
        assert Loadout.US_OFFICER_NCO.items[0].weapon == Weapon.M1_GARAND
        assert Loadout.US_OFFICER_NCO.items[1].weapon == Weapon.MK2_GRENADE
        assert Loadout.US_OFFICER_NCO.items[2].weapon is None



import os
from pathlib import Path
from typing import TypedDict

from dotenv import load_dotenv

ENV_PATH = Path(".env")
if ENV_PATH.exists():
    load_dotenv(ENV_PATH)

_hll_port = os.getenv("HLL_PORT")

HLL_HOST = os.getenv("HLL_HOST")
HLL_PORT = int(_hll_port) if _hll_port else None
HLL_PASSWORD = os.getenv("HLL_PASSWORD")


class HLLServerCredentials(TypedDict):
    host: str
    port: int
    password: str


HLL_SERVER_CREDENTIALS: HLLServerCredentials | None = (
    {
        "host": HLL_HOST,
        "port": HLL_PORT,
        "password": HLL_PASSWORD,
    }
    if (HLL_HOST and HLL_PORT and HLL_PASSWORD)
    else None
)
